# pylint:disable=no-self-use
from __future__ import annotations
from collections import defaultdict
from typing import TYPE_CHECKING
import logging

import networkx
import capstone

from pyvex.stmt import Put
from pyvex.expr import RdTmp
import angr.ailment as ailment

from angr.code_location import ExternalCodeLocation

from angr.calling_conventions import (
    SimFunctionArgument,
    SimRegArg,
    SimStackArg,
    SimCC,
    default_cc,
    SimCCMicrosoftThiscall,
)
from angr.errors import SimTranslationError
from angr.sim_type import (
    SimTypeCppFunction,
    SimTypeInt,
    SimTypeFunction,
    SimType,
    SimTypeLongLong,
    SimTypeShort,
    SimTypeChar,
    SimTypeBottom,
    SimTypeFloat,
    SimTypeDouble,
    parse_cpp_file,
)
from angr.sim_variable import SimStackVariable, SimRegisterVariable
from angr.knowledge_plugins.key_definitions.atoms import Register, MemoryLocation, SpOffset
from angr.knowledge_plugins.key_definitions.tag import ReturnValueTag
from angr.knowledge_plugins.key_definitions.constants import OP_BEFORE, OP_AFTER
from angr.knowledge_plugins.key_definitions.rd_model import ReachingDefinitionsModel
from angr.knowledge_plugins.variables.variable_access import VariableAccessSort
from angr.knowledge_plugins.functions import Function
from angr.utils.constants import DEFAULT_STATEMENT
from angr.utils.ssa import get_reg_offset_base_and_size, get_reg_offset_base
from angr import SIM_PROCEDURES
from angr.analyses import Analysis, register_analysis, ReachingDefinitionsAnalysis
from angr.analyses.reaching_definitions import get_all_definitions
from .utils import is_sane_register_variable

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg import CFGModel
    from angr.knowledge_plugins.key_definitions.uses import Uses
    from angr.knowledge_plugins.key_definitions.definition import Definition

l = logging.getLogger(name=__name__)


class CallSiteFact:
    """
    Store facts about each call site.
    """

    def __init__(self, return_value_used):
        self.return_value_used: bool = return_value_used
        self.args = []


class UpdateArgumentsOption:
    """
    Enums for controlling the argument updating behavior in _adjust_cc.
    """

    DoNotUpdate = 0
    AlwaysUpdate = 1
    UpdateWhenCCHasNoArgs = 2


class CallingConventionAnalysis(Analysis):
    """
    Analyze the calling convention of a function and guess a probable prototype.

    The calling convention of a function can be inferred at both its call sites and the function itself. At call sites,
    we consider all register and stack variables that are not alive after the function call as parameters to this
    function. In the function itself, we consider all register and stack variables that are read but without
    initialization as parameters. Then we synthesize the information from both locations and make a reasonable
    inference of calling convention of this function.

    :ivar _function:    The function to recover calling convention for.
    :ivar _variable_manager:    A handy accessor to the variable manager.
    :ivar _cfg:         A reference of the CFGModel of the current binary. It is used to discover call sites of the
                        current function in order to perform analysis at call sites.
    :ivar analyze_callsites:    True if we should analyze all call sites of the current function to determine the
                                calling convention and arguments. This can be time-consuming if there are many call
                                sites to analyze.
    :ivar cc:           The recovered calling convention for the function.
    :ivar _collect_facts:       True if we should run FunctionFactCollector to collect input arguments and return
                                value size. False if input arguments and return value size are provided by the user.
    """

    def __init__(
        self,
        func: Function | int | str | None,
        cfg: CFGModel | None = None,
        analyze_callsites: bool = False,
        caller_func_addr: int | None = None,
        callsite_block_addr: int | None = None,
        callsite_insn_addr: int | None = None,
        func_graph: networkx.DiGraph | None = None,
        input_args: list[SimRegArg | SimStackArg] | None = None,
        retval_size: int | None = None,
        collect_facts: bool = False,
    ):
        if func is not None and not isinstance(func, Function):
            func = self.kb.functions[func]
        self._function = func
        self._variable_manager = self.kb.variables
        self._cfg = cfg
        self.analyze_callsites = analyze_callsites
        self.caller_func_addr = caller_func_addr
        self.callsite_block_addr = callsite_block_addr
        self.callsite_insn_addr = callsite_insn_addr
        self._func_graph = func_graph
        self._input_args = input_args
        self._retval_size = retval_size
        self._collect_facts = collect_facts

        if self._retval_size is not None and self._input_args is None:
            # retval size will be ignored if input_args is not specified - user error?
            raise TypeError(
                "input_args must be provided to use retval_size. Otherwise please set both input_args and "
                "retval_size to None."
            )

        self.cc: SimCC | None = None
        self.prototype: SimTypeFunction | None = None
        self.prototype_libname: str | None = None
        self.proto_from_symbol: bool = False

        if self._cfg is None and "CFGFast" in self.kb.cfgs:
            self._cfg = self.kb.cfgs["CFGFast"]

        if self._function is not None:
            # caller function analysis mode
            self._analyze()
        elif (
            self.analyze_callsites
            and self.caller_func_addr is not None
            and self.callsite_block_addr is not None
            and self.callsite_insn_addr is not None
        ):
            # callsite analysis mode
            self._analyze_callsite_only()
        else:
            raise TypeError(
                'You must specify a function to analyze, or specify "caller_func_addr",'
                ' "callsite_block_addr" and "callsite_insn_addr" to only analyze a call site.'
            )

        if self.prototype is not None:
            self.prototype = self.prototype.with_arch(self.project.arch)

    def _analyze(self):
        """
        The major analysis routine.
        """

        assert self._function is not None

        demangled_name = self._function.demangled_name
        if demangled_name != self._function.name:
            r_demangled = self._analyze_demangled_name(demangled_name)
            if r_demangled is not None:
                self.cc, self.prototype, self.prototype_libname = r_demangled
                self.proto_from_symbol = True
                return

        if self._function.is_simprocedure:
            hooker = self.project.hooked_by(self._function.addr)
            if isinstance(
                hooker,
                (
                    SIM_PROCEDURES["stubs"]["UnresolvableCallTarget"],
                    SIM_PROCEDURES["stubs"]["UnresolvableJumpTarget"],
                    SIM_PROCEDURES["stubs"]["UserHook"],
                ),
            ):
                return

            if (
                hooker is not None
                and hooker.cc is not None
                and hooker.is_function
                and not hooker.guessed_prototype
                and hooker.prototype is not None
            ):
                # copy the calling convention and prototype from the SimProcedure instance
                self.cc = hooker.cc
                self.prototype = hooker.prototype
                self.prototype_libname = hooker.library_name
                self.proto_from_symbol = True
                return

            if self._function.prototype is None:
                # try our luck
                # we set ignore_binary_name to True because the binary name SimProcedures is "cle##externs" and does not
                # match any library name
                self._function.find_declaration(ignore_binary_name=True)

            self.cc = self._function.calling_convention
            self.prototype = self._function.prototype
            self.prototype_libname = self._function.prototype_libname

            if self.cc is None or self.prototype is None:
                for include_callsite_preds in [False, True]:
                    callsite_facts = self._extract_and_analyze_callsites(
                        max_analyzing_callsites=1,
                        include_callsite_preds=include_callsite_preds,
                    )
                    cc_cls = default_cc(
                        self.project.arch.name,
                        platform=(
                            self.project.simos.name
                            if self.project is not None and self.project.simos is not None
                            else None
                        ),
                    )
                    cc = cc_cls(self.project.arch) if cc_cls is not None else None
                    prototype = None
                    if callsite_facts:
                        if self.prototype is None:
                            proto = SimTypeFunction([], SimTypeBottom(label="void"))
                        else:
                            proto = self.prototype
                        prototype = self._adjust_prototype(
                            proto,
                            callsite_facts,
                            update_arguments=UpdateArgumentsOption.AlwaysUpdate,
                        )
                        if prototype.args:
                            break
                self.cc = cc  # type: ignore
                self.prototype = prototype  # type: ignore
            return
        if self._function.is_plt:
            r_plt = self._analyze_plt()
            if r_plt is not None:
                self.cc, self.prototype, self.prototype_libname, proto_guessed = r_plt
                self.proto_from_symbol = not proto_guessed
            return

        # we gotta analyze the function properly
        if self._collect_facts and self._input_args is None and self._retval_size is None:
            facts = self.project.analyses.FunctionFactCollector(self._function, kb=self.kb)
            self._input_args = facts.input_args
            self._retval_size = facts.retval_size

        r = self._analyze_function()
        if r is None:
            l.warning("Cannot determine calling convention for %r.", self._function)
        else:
            # adjust prototype if needed
            cc, prototype = r
            if self.analyze_callsites:
                # only take the first 3 because running reaching definition analysis on all functions is costly
                callsite_facts = self._extract_and_analyze_callsites(max_analyzing_callsites=3)
                prototype = (
                    self._adjust_prototype(
                        prototype, callsite_facts, update_arguments=UpdateArgumentsOption.UpdateWhenCCHasNoArgs
                    )
                    if prototype is not None
                    else None
                )

            self.cc = cc
            self.prototype = prototype

    def _analyze_callsite_only(self):
        assert self.caller_func_addr is not None
        assert self.callsite_block_addr is not None
        assert self.callsite_insn_addr is not None
        cc, prototype = None, None

        for include_callsite_preds in [False, True]:
            fact = self._analyze_callsite(
                self.caller_func_addr,
                self.callsite_block_addr,
                self.callsite_insn_addr,
                include_preds=include_callsite_preds,
            )
            if fact is None:
                continue
            callsite_facts = [fact]
            cc_cls = default_cc(
                self.project.arch.name,
                platform=(
                    self.project.simos.name if self.project is not None and self.project.simos is not None else None
                ),
            )
            cc = cc_cls(self.project.arch) if cc_cls is not None else None
            prototype = SimTypeFunction([], None)
            prototype = self._adjust_prototype(
                prototype, callsite_facts, update_arguments=UpdateArgumentsOption.AlwaysUpdate
            )
            if prototype.args:
                break

        self.cc = cc
        self.prototype = prototype

    def _analyze_plt(self) -> tuple[SimCC, SimTypeFunction | None, str | None, bool | None] | None:
        """
        Get the calling convention for a PLT stub.

        :return:    A calling convention, the function type, as well as the library name if available.
        """
        assert self._function is not None

        if len(self._function.jumpout_sites) != 1:
            l.warning(
                "%r has more than one jumpout sites. It does not look like a PLT stub. Please report to GitHub.",
                self._function,
            )
            return None

        jo_site = self._function.jumpout_sites[0]

        successors = list(self._function.transition_graph.successors(jo_site))
        if len(successors) != 1:
            l.warning(
                "%r has more than one successors. It does not look like a PLT stub. Please report to GitHub.",
                self._function,
            )
            return None

        try:
            real_func = self.kb.functions.get_by_addr(successors[0].addr)
        except KeyError:
            # the real function does not exist for some reason
            real_func = None

        if real_func is not None:
            if real_func.calling_convention is None:
                cc_cls = default_cc(self.project.arch.name)
                if cc_cls is None:
                    # can't determine the default calling convention for this architecture
                    return None
                cc = cc_cls(self.project.arch)
            else:
                cc = real_func.calling_convention
            if real_func.is_simprocedure:
                if self.project.is_hooked(real_func.addr):
                    # prioritize the hooker
                    hooker = self.project.hooked_by(real_func.addr)
                    if hooker is not None and hooker.is_function and not hooker.guessed_prototype:
                        # we only take the prototype from the SimProcedure if
                        # - the SimProcedure is a function
                        # - the prototype of the SimProcedure is not guessed
                        return cc, hooker.prototype, hooker.library_name, False
                if real_func.prototype is not None:
                    return cc, real_func.prototype, real_func.prototype_libname, False
            else:
                return cc, real_func.prototype, real_func.prototype_libname, False

        if self.analyze_callsites:
            # determine the calling convention by analyzing its callsites
            callsite_facts = self._extract_and_analyze_callsites(max_analyzing_callsites=1)
            cc_cls = default_cc(self.project.arch.name)
            if cc_cls is None:
                # can't determine the default calling convention for this architecture
                return None
            cc = cc_cls(self.project.arch)
            prototype = SimTypeFunction([], None)
            prototype = self._adjust_prototype(
                prototype, callsite_facts, update_arguments=UpdateArgumentsOption.AlwaysUpdate
            )
            return cc, prototype, None, True

        return None

    def _analyze_demangled_name(self, name: str) -> tuple[SimCC, SimTypeFunction, str | None] | None:
        """
        Analyze a function with a demangled name. Only C++ names are supported for now.

        :param name:    The demangled name of the function.
        :return:        A tuple of the calling convention, the function type, and the library name if available.
        """
        parsed, _ = parse_cpp_file(name)
        if not parsed or len(parsed) != 1:
            return None
        proto = next(iter(parsed.values()))
        if (
            isinstance(proto, SimTypeCppFunction)
            and self.project.simos.name == "Win32"
            and self.project.arch.name == "X86"
            and proto.convention == "__thiscall"
        ):
            cc_cls = SimCCMicrosoftThiscall
        else:
            cc_cls = default_cc(self.project.arch.name, self.project.simos.name)
            assert cc_cls is not None
        cc = cc_cls(self.project.arch)
        return cc, proto, None

    def _analyze_function(self) -> tuple[SimCC, SimTypeFunction] | None:
        """
        Go over the variable information in variable manager for this function, and return all uninitialized
        register/stack variables.
        """
        assert self._function is not None

        if self._function.is_simprocedure or self._function.is_plt:
            # we do not analyze SimProcedures or PLT stubs
            return None

        if self._input_args is None:
            if not self._variable_manager.has_function_manager(self._function.addr):
                l.warning("Please run variable recovery on %r before analyzing its calling convention.", self._function)
                return None
            vm = self._variable_manager[self._function.addr]
            retval_size = vm.ret_val_size
            input_variables = vm.input_variables()
            input_args = self._args_from_vars(input_variables, vm)
        else:
            input_args = set(self._input_args)
            retval_size = self._retval_size

        # check if this function is a variadic function
        if self.project.arch.name == "AMD64":
            is_variadic, fixed_args = self.is_va_start_amd64(self._function)
        else:
            is_variadic = False
            fixed_args = None

        # TODO: properly determine sp_delta
        sp_delta = self.project.arch.bytes if self.project.arch.call_pushes_ret else 0

        full_input_args = self._consolidate_input_args(input_args)
        full_input_args_copy = list(full_input_args)  # input_args might be modified by find_cc()
        cc = SimCC.find_cc(self.project.arch, full_input_args_copy, sp_delta, platform=self.project.simos.name)

        # update input_args according to the difference between full_input_args and full_input_args_copy
        for a in full_input_args:
            if a not in full_input_args_copy and a in input_args:
                input_args.remove(a)

        if cc is None:
            l.warning(
                "_analyze_function(): Cannot find a calling convention for %r that fits the given arguments.",
                self._function,
            )
            return None
        # reorder args
        args = self._reorder_args(input_args, cc)
        if fixed_args is not None:
            args = args[:fixed_args]

        # guess the type of the return value -- it's going to be a wild guess...
        ret_type = self._guess_retval_type(cc, retval_size)
        if self._function.name == "main" and self.project.arch.bits == 64 and isinstance(ret_type, SimTypeLongLong):
            # hack - main must return an int even in 64-bit binaries
            ret_type = SimTypeInt()
        prototype = SimTypeFunction([self._guess_arg_type(arg, cc) for arg in args], ret_type, variadic=is_variadic)

        return cc, prototype

    def _analyze_callsite(
        self,
        caller_addr: int,
        caller_block_addr: int,
        call_insn_addr: int,
        include_preds: bool = False,
    ) -> CallSiteFact | None:
        func = self.kb.functions[caller_addr]
        subgraph = self._generate_callsite_subgraph(func, caller_block_addr, include_preds=include_preds)
        if subgraph is None:
            # failed to generate a subgraph when the caller block cannot be found in the function graph
            return None

        observation_points: list = [("insn", call_insn_addr, OP_BEFORE), ("node", caller_block_addr, OP_AFTER)]

        # find the return site
        caller_block = next(iter(bb for bb in subgraph if bb.addr == caller_block_addr))
        return_site_block = next(iter(subgraph.successors(caller_block)), None)
        if return_site_block is not None:
            observation_points.append(("node", return_site_block.addr, OP_AFTER))

        rda = self.project.analyses[ReachingDefinitionsAnalysis].prep()(
            func,
            func_graph=subgraph,
            observation_points=observation_points,
        )
        # rda_model: Optional[ReachingDefinitionsModel] = self.kb.defs.get_model(caller.addr)
        return self._collect_callsite_fact(caller_block, call_insn_addr, rda.model)

    def _extract_and_analyze_callsites(
        self,
        max_analyzing_callsites: int = 3,
        include_callsite_preds: bool = False,
    ) -> list[CallSiteFact]:  # pylint:disable=no-self-use
        """
        Analyze all call sites of the function and determine the possible number of arguments and if the function
        returns anything or not.
        """

        assert self._function is not None

        if self._cfg is None:
            l.warning("CFG is not provided. Skip calling convention analysis at call sites.")
            return []

        node = self._cfg.get_any_node(self._function.addr)
        if node is None:
            l.warning("%r is not in the CFG. Skip calling convention analysis at call sites.", self._function)

        facts = []
        in_edges = self._cfg.graph.in_edges(node, data=True)

        call_sites_by_function: dict[Function, list[tuple[int, int]]] = defaultdict(list)

        if len(in_edges) == 1:
            src, _, data = next(iter(in_edges))
            if (
                data.get("jumpkind", "Ijk_Call") == "Ijk_Boring"
                and self.kb.functions.contains_addr(src.function_address)
                and self.kb.functions[src.function_address].is_plt
            ):
                # find callers to the PLT stub instead
                in_edges = self._cfg.graph.in_edges(src, data=True)

        for src, _, data in sorted(in_edges, key=lambda x: x[0].addr):
            edge_type = data.get("jumpkind", "Ijk_Call")
            if not (edge_type == "Ijk_Call" or (edge_type == "Ijk_Boring" and self._cfg.graph.out_degree[src] == 1)):
                continue
            if not self.kb.functions.contains_addr(src.function_address):
                continue
            caller = self.kb.functions[src.function_address]
            if caller.is_simprocedure or caller.is_alignment:
                # do not analyze SimProcedures or alignment stubs
                continue
            if src.instruction_addrs:
                call_sites_by_function[caller].append((src.addr, src.instruction_addrs[-1]))

        call_sites_by_function_list = sorted(call_sites_by_function.items(), key=lambda x: x[0].addr)[
            :max_analyzing_callsites
        ]
        ctr = 0

        for caller, call_site_tuples in call_sites_by_function_list:
            if ctr >= max_analyzing_callsites:
                break

            # generate a subgraph that only contains the basic block that does the call and the basic block after the
            # call.
            for call_site_tuple in call_site_tuples:
                caller_block_addr, call_insn_addr = call_site_tuple
                fact = self._analyze_callsite(
                    caller.addr,
                    caller_block_addr,
                    call_insn_addr,
                    include_preds=include_callsite_preds,
                )
                if fact is None:
                    continue
                facts.append(fact)

                ctr += 1
                if ctr >= max_analyzing_callsites:
                    break

        return facts

    def _generate_callsite_subgraph(
        self,
        func: Function,
        callsite_block_addr: int,
        include_preds: bool = False,
    ) -> networkx.DiGraph | None:
        func_graph = self._func_graph if self._func_graph is not None else func.graph

        the_block = next(iter(nn for nn in func_graph if nn.addr == callsite_block_addr), None)
        if the_block is None:
            return None

        subgraph = networkx.DiGraph()
        subgraph.add_node(the_block)

        if include_preds:
            # add a predecessor
            for src, _, data in func_graph.in_edges(the_block, data=True):
                if src is not the_block:
                    subgraph.add_edge(src, the_block, **data)
                    break  # only add the first non-cycle in-edge

        for _, dst, data in func_graph.out_edges(the_block, data=True):
            subgraph.add_edge(the_block, dst, **data)

            # If the target block contains only direct jump statements and has only one successor,
            # include its successor.

            # Re-lift the target block
            dst_block_size = func.get_block_size(dst.addr)
            if dst_block_size is not None and dst_block_size > 0:
                dst_bb = self.project.factory.block(dst.addr, dst_block_size, opt_level=1)
                try:
                    vex_block = dst_bb.vex
                except SimTranslationError:
                    # failed to lift the block
                    continue

                # If there is only one 'IMark' statement in vex --> the target block contains only direct jump
                if (
                    len(vex_block.statements) == 1
                    and vex_block.statements[0].tag == "Ist_IMark"
                    and func.graph.out_degree(dst) == 1
                ):
                    for _, jmp_dst, jmp_data in func_graph.out_edges(dst, data=True):
                        subgraph.add_edge(dst, jmp_dst, **jmp_data)

        return subgraph

    def _collect_callsite_fact(
        self,
        caller_block,
        call_insn_addr: int,
        rda: ReachingDefinitionsModel,
    ) -> CallSiteFact:
        fact = CallSiteFact(
            True,  # by default we treat all return values as used
        )

        default_cc_cls = default_cc(
            self.project.arch.name,
            platform=self.project.simos.name if self.project is not None and self.project.simos is not None else None,
        )
        if default_cc_cls is not None:
            cc: SimCC = default_cc_cls(self.project.arch)
            self._analyze_callsite_return_value_uses(cc, caller_block.addr, rda, fact)
            self._analyze_callsite_arguments(cc, caller_block, call_insn_addr, rda, fact)

        return fact

    def _analyze_callsite_return_value_uses(
        self, cc: SimCC, caller_block_addr: int, rda: ReachingDefinitionsModel, fact: CallSiteFact
    ) -> None:
        all_defs: set[Definition] = {
            def_
            for def_ in rda.all_uses._uses_by_definition
            if (
                (def_.codeloc.block_addr == caller_block_addr and def_.codeloc.stmt_idx == DEFAULT_STATEMENT)
                or any(isinstance(tag, ReturnValueTag) for tag in def_.tags)
            )
        }
        all_uses: Uses = rda.all_uses

        # determine if the return value is used
        return_val = cc.RETURN_VAL
        if return_val is not None and isinstance(return_val, SimRegArg):
            return_reg_offset, _ = self.project.arch.registers[return_val.reg_name]

            # find the def of the return val
            try:
                return_def = next(
                    iter(d for d in all_defs if isinstance(d.atom, Register) and d.atom.reg_offset == return_reg_offset)
                )
            except StopIteration:
                return_def = None
                fact.return_value_used = False

            if return_def is not None:
                # is it used?
                uses = all_uses.get_uses(return_def)
                if uses:
                    # the return value is used!
                    fact.return_value_used = True
                else:
                    fact.return_value_used = False

    def _analyze_callsite_arguments(
        self,
        cc: SimCC,
        caller_block,
        call_insn_addr: int,
        rda: ReachingDefinitionsModel,
        fact: CallSiteFact,
    ) -> None:
        # determine if potential register and stack arguments are set
        observation_key = "insn", call_insn_addr, OP_BEFORE
        state = rda.observed_results.get(observation_key)
        if state is None:
            # the observation state is not found. it can happen if call_insn_addr is incorrect, which may happen (but
            # rarely) on incorrect CFGs.
            return

        defs_by_reg_offset: dict[int, list[Definition]] = defaultdict(list)
        all_reg_defs: set[Definition] = get_all_definitions(state.registers)
        all_stack_defs: set[Definition] = get_all_definitions(state.stack)
        for d in all_reg_defs:
            if (
                isinstance(d.atom, Register)
                and not isinstance(d.codeloc, ExternalCodeLocation)
                and not (d.codeloc.block_addr == caller_block.addr and d.codeloc.stmt_idx == DEFAULT_STATEMENT)
            ):
                # do an extra check because of how entry and callN work on Xtensa
                if isinstance(caller_block, ailment.Block) and self._likely_saving_temp_reg(
                    caller_block, d, all_reg_defs
                ):
                    continue
                defs_by_reg_offset[d.offset].append(d)
        defined_reg_offsets = set(defs_by_reg_offset.keys())
        sp_offset = 0
        if self.project.arch.bits in {32, 64}:
            # Calculate the offsets between sp and stack defs
            sp_offset = state.get_sp_offset()
            if sp_offset is None:
                # We can not find the sp_offset when sp is concrete
                # e.g.,
                # LDR     R2, =0x20070000
                # STR     R1, [R3,#0x38]
                # MOV     SP, R2
                # In this case, just assume sp_offset = 0
                sp_offset = 0
        defs_by_stack_offset = {
            d.atom.addr.offset - sp_offset: d
            for d in all_stack_defs
            if isinstance(d.atom, MemoryLocation) and isinstance(d.atom.addr, SpOffset)
        }

        default_type_cls = SimTypeInt if self.project.arch.bits == 32 else SimTypeLongLong
        arg_session = cc.arg_session(default_type_cls().with_arch(self.project.arch))
        temp_args: list[SimFunctionArgument | None] = []
        expected_args: list[SimFunctionArgument] = []
        for _ in range(30):  # at most 30 arguments
            arg_loc = cc.next_arg(arg_session, default_type_cls().with_arch(self.project.arch))
            expected_args.append(arg_loc)
            if isinstance(arg_loc, SimRegArg):
                reg_offset = self.project.arch.registers[arg_loc.reg_name][0]
                # is it initialized?
                if reg_offset in defined_reg_offsets:
                    temp_args.append(arg_loc)
                else:
                    # no more arguments
                    temp_args.append(None)
            elif isinstance(arg_loc, SimStackArg):
                if arg_loc.stack_offset - cc.STACKARG_SP_DIFF in defs_by_stack_offset:
                    temp_args.append(arg_loc)
                else:
                    # no more arguments
                    break
            else:
                break

        if None in temp_args:
            # we be very conservative here and ignore all arguments starting from the first missing one
            first_none_idx = temp_args.index(None)
            fact.args = temp_args[:first_none_idx]
        else:
            fact.args = temp_args

    def _adjust_prototype(
        self,
        proto: SimTypeFunction,
        facts: list[CallSiteFact],
        update_arguments: int = UpdateArgumentsOption.DoNotUpdate,
    ) -> SimTypeFunction:
        # is the return value used anywhere?
        if facts:
            if all(fact.return_value_used is False for fact in facts):
                proto.returnty = SimTypeBottom(label="void")
            else:
                if proto.returnty is None or isinstance(proto.returnty, SimTypeBottom):
                    returnty = {32: SimTypeInt, 16: SimTypeShort, 64: SimTypeLongLong}.get(
                        self.project.arch.bits, SimTypeInt
                    )(signed=True)
                    proto.returnty = returnty.with_arch(self.project.arch)

        if (
            update_arguments == UpdateArgumentsOption.AlwaysUpdate
            or (update_arguments == UpdateArgumentsOption.UpdateWhenCCHasNoArgs and not proto.args)
        ) and len({len(fact.args) for fact in facts}) == 1:
            fact = next(iter(facts))
            proto.args = tuple(
                self._guess_arg_type(arg) if arg is not None else SimTypeInt().with_arch(self.project.arch)
                for arg in fact.args
            )

        return proto

    def _args_from_vars(self, variables: list, var_manager):
        """
        Derive function arguments from input variables.

        :param variables:
        :param var_manager: The variable manager of this function.
        :return:
        """

        assert self._function is not None

        args = set()
        ret_addr_offset = 0 if not self.project.arch.call_pushes_ret else self.project.arch.bytes

        reg_vars_with_single_access: list[SimRegisterVariable] = []

        def_cc = default_cc(
            self.project.arch.name,
            platform=self.project.simos.name if self.project is not None and self.project.simos is not None else None,
        )
        for variable in variables:
            if isinstance(variable, SimStackVariable):
                # a stack variable. convert it to a stack argument.
                # TODO: deal with the variable base
                if self.project.arch.call_pushes_ret and variable.offset <= 0:
                    # skip the return address on the stack
                    # TODO: make sure it was the return address
                    continue
                if variable.offset - ret_addr_offset >= 0:
                    arg = SimStackArg(variable.offset - ret_addr_offset, variable.size)
                    args.add(arg)
            elif isinstance(variable, SimRegisterVariable):
                # a register variable, convert it to a register argument
                if not is_sane_register_variable(self.project.arch, variable.reg, variable.size, def_cc=def_cc):
                    continue
                reg_name = self.project.arch.translate_register_name(variable.reg, size=variable.size)
                arg = SimRegArg(reg_name, variable.size)
                args.add(arg)

                accesses = var_manager.get_variable_accesses(variable)
                if len(accesses) == 1:
                    reg_vars_with_single_access.append(variable)
            else:
                l.error("Unsupported type of variable %s.", type(variable))

        # the function might be saving registers at the beginning and restoring them at the end
        # we should remove all registers that are strictly callee-saved and are not used anywhere in this function
        end_blocks = [(endpoint.addr, endpoint.size) for endpoint in self._function.endpoints_with_type["return"]]

        restored_reg_vars: set[SimRegArg] = set()

        # is there any instruction that restores this register in any end blocks?
        if reg_vars_with_single_access:
            if self._function.returning is False:
                # no restoring is required if this function does not return
                for var_ in reg_vars_with_single_access:
                    reg_name = self.project.arch.translate_register_name(var_.reg, size=var_.size)
                    restored_reg_vars.add(SimRegArg(reg_name, var_.size))

            else:
                reg_offsets: set[int] = {r.reg for r in reg_vars_with_single_access}
                for var_ in var_manager.get_variables(sort="reg"):
                    if var_.reg in (reg_offsets - {self.project.arch.ret_offset}):
                        # check if there is only a write to it
                        accesses = var_manager.get_variable_accesses(var_)
                        if len(accesses) == 1 and accesses[0].access_type == VariableAccessSort.WRITE:
                            found = False
                            for end_block_addr, end_block_size in end_blocks:
                                if end_block_addr <= accesses[0].location.ins_addr < end_block_addr + end_block_size:
                                    found = True
                                    break

                            if found:
                                reg_name = self.project.arch.translate_register_name(var_.reg, size=var_.size)
                                restored_reg_vars.add(SimRegArg(reg_name, var_.size))

        return args.difference(restored_reg_vars)

    def _consolidate_input_args(self, input_args: set[SimRegArg | SimStackArg]) -> set[SimRegArg | SimStackArg]:
        """
        Consolidate register arguments by converting partial registers to full registers on certain architectures.

        :param input_args:  A set of input arguments.
        :return:            A set of consolidated input args.
        """

        if self.project.arch.name in {"AMD64", "X86"}:
            new_input_args = set()
            for a in input_args:
                if isinstance(a, SimRegArg) and a.size < self.project.arch.bytes:
                    # use complete registers on AMD64 and X86
                    reg_offset, reg_size = self.project.arch.registers[a.reg_name]
                    full_reg_offset, full_reg_size = get_reg_offset_base_and_size(
                        reg_offset, self.project.arch, size=reg_size
                    )
                    full_reg_name = self.project.arch.translate_register_name(full_reg_offset, size=full_reg_size)
                    arg = SimRegArg(full_reg_name, full_reg_size)
                    if arg not in new_input_args:
                        new_input_args.add(arg)
                else:
                    new_input_args.add(a)
            return new_input_args

        return input_args

    def _reorder_args(self, args: set[SimRegArg | SimStackArg], cc: SimCC) -> list[SimRegArg | SimStackArg]:
        """
        Reorder arguments according to the calling convention identified.

        :param args:   A set of arguments that haven't been ordered.
        :param cc:    The identified calling convention.
        :return:            A reordered list of args.
        """

        def _is_same_reg(rn0: str, rn1: str) -> bool:
            """
            Check if rn0 and rn1 belong to the same base register.

            :param rn0:     Register name of the first register.
            :param rn1:     Register name of the second register.
            :return:        True if they belong to the same base register; False otherwise.
            """
            if rn0 == rn1:
                return True
            off0, sz0 = self.project.arch.registers[rn0]
            full_off0 = get_reg_offset_base(off0, self.project.arch, sz0)
            off1, sz1 = self.project.arch.registers[rn1]
            full_off1 = get_reg_offset_base(off1, self.project.arch, sz1)
            return full_off0 == full_off1

        reg_args = []

        # split args into two lists
        int_args = []
        fp_args = []
        for arg in args:
            if isinstance(arg, SimRegArg):
                if cc.FP_ARG_REGS and arg.reg_name in cc.FP_ARG_REGS:
                    fp_args.append(arg)
                else:
                    int_args.append(arg)

        initial_stack_args = sorted([a for a in args if isinstance(a, SimStackArg)], key=lambda a: a.stack_offset)
        # ensure stack args are consecutive if necessary
        if cc.STACKARG_SP_DIFF is not None and initial_stack_args:
            arg_by_offset = {a.stack_offset: a for a in initial_stack_args}
            init_stackarg_offset = cc.STACKARG_SP_DIFF + cc.STACKARG_SP_BUFF
            int_arg_size = self.project.arch.bytes
            for stackarg_offset in range(init_stackarg_offset, max(arg_by_offset), int_arg_size):
                if stackarg_offset not in arg_by_offset:
                    arg_by_offset[stackarg_offset] = SimStackArg(stackarg_offset, int_arg_size)
            stack_args = [arg_by_offset[offset] for offset in sorted(arg_by_offset)]
        else:
            stack_args = initial_stack_args

        stack_int_args = [a for a in stack_args if not a.is_fp]
        stack_fp_args = [a for a in stack_args if a.is_fp]
        # match int args first
        for reg_name in cc.ARG_REGS:
            try:
                arg = next(iter(a for a in int_args if isinstance(a, SimRegArg) and _is_same_reg(a.reg_name, reg_name)))
            except StopIteration:
                # have we reached the end of the args list?
                if [a for a in int_args if isinstance(a, SimRegArg)] or len(stack_int_args) > 0:
                    # haven't reached the end yet or there are stack args
                    arg = SimRegArg(reg_name, self.project.arch.bytes)
                else:
                    break
            reg_args.append(arg)
            if arg in int_args:
                int_args.remove(arg)

        # match fp args later
        if fp_args:
            for reg_name in cc.FP_ARG_REGS:
                try:
                    arg = next(
                        iter(a for a in fp_args if isinstance(a, SimRegArg) and _is_same_reg(a.reg_name, reg_name))
                    )
                except StopIteration:
                    # have we reached the end of the args list?
                    if [a for a in fp_args if isinstance(a, SimRegArg)] or len(stack_fp_args) > 0:
                        # haven't reached the end yet or there are stack args
                        arg = SimRegArg(reg_name, self.project.arch.bytes)
                    else:
                        break
                reg_args.append(arg)
                if arg in fp_args:
                    fp_args.remove(arg)

        return reg_args + int_args + fp_args + stack_args

    def _guess_arg_type(self, arg: SimFunctionArgument, cc: SimCC | None = None) -> SimType:
        if cc is not None and cc.FP_ARG_REGS and isinstance(arg, SimRegArg) and arg.reg_name in cc.FP_ARG_REGS:
            if arg.size == 4:
                return SimTypeFloat()
            if arg.size == 8:
                return SimTypeDouble()

        if arg.size == 4:
            return SimTypeInt()
        if arg.size == 8:
            return SimTypeLongLong()
        if arg.size == 2:
            return SimTypeShort()
        if arg.size == 1:
            return SimTypeChar()
        # Unsupported for now
        return SimTypeBottom()

    def _guess_retval_type(self, cc: SimCC, ret_val_size: int | None) -> SimType:
        assert self._function is not None

        if cc.FP_RETURN_VAL and self._function.ret_sites:
            # examine the last block of the function and see which registers are assigned to
            for ret_block in self._function.ret_sites:
                fpretval_updated, retval_updated = False, False
                fp_reg_size = 0
                try:
                    irsb = self.project.factory.block(ret_block.addr, size=ret_block.size).vex
                except SimTranslationError:
                    # failed to lift the block
                    continue
                for stmt in irsb.statements:
                    if isinstance(stmt, Put) and isinstance(stmt.data, RdTmp):
                        reg_size = irsb.tyenv.sizeof(stmt.data.tmp) // self.project.arch.byte_width  # type: ignore
                        reg_name = self.project.arch.translate_register_name(stmt.offset, size=reg_size)
                        if isinstance(cc.FP_RETURN_VAL, SimRegArg) and reg_name == cc.FP_RETURN_VAL.reg_name:
                            fpretval_updated = True
                            fp_reg_size = reg_size
                        elif isinstance(cc.RETURN_VAL, SimRegArg) and reg_name == cc.RETURN_VAL.reg_name:
                            retval_updated = True

                if fpretval_updated and not retval_updated:
                    # possibly float
                    return SimTypeFloat() if fp_reg_size == 4 else SimTypeDouble()

        if ret_val_size is not None:
            if ret_val_size == 1:
                return SimTypeChar()
            if ret_val_size == 2:
                return SimTypeShort()
            if 3 <= ret_val_size <= 4:
                return SimTypeInt()
            if 5 <= ret_val_size <= 8:
                return SimTypeLongLong()

        return SimTypeBottom(label="void")

    @staticmethod
    def _likely_saving_temp_reg(ail_block: ailment.Block, d: Definition, all_reg_defs: set[Definition]) -> bool:
        if (
            d.codeloc.block_addr == ail_block.addr
            and d.codeloc.stmt_idx is not None
            and d.codeloc.stmt_idx < len(ail_block.statements)
        ):
            stmt = ail_block.statements[d.codeloc.stmt_idx]
            if isinstance(stmt, ailment.Stmt.Assignment) and isinstance(stmt.src, ailment.Expr.Register):
                src_offset = stmt.src.reg_offset
                src_reg_def = next(
                    iter(
                        d_ for d_ in all_reg_defs if isinstance(d_.atom, Register) and d_.atom.reg_offset == src_offset
                    ),
                    None,
                )
                if src_reg_def is not None and isinstance(src_reg_def.codeloc, ExternalCodeLocation):
                    return True
        return False

    def is_va_start_amd64(self, func: Function) -> tuple[bool, int | None]:
        # TODO: Use a better pattern matching approach
        if len(func.block_addrs_set) < 3:
            return False, None
        if func.startpoint is None:
            return False, None

        head = func.startpoint
        out_edges = list(func.transition_graph.out_edges(head, data=True))
        if len(out_edges) != 2:
            return False, None
        succ0, succ1 = out_edges[0][1], out_edges[1][1]
        if func.transition_graph.has_edge(succ0, succ1):
            mid = succ0
        elif func.transition_graph.has_edge(succ1, succ0):
            mid = succ1
        else:
            return False, None

        # compare instructions
        for insn in self.project.factory.block(mid.addr, size=mid.size).capstone.insns:
            if insn.mnemonic != "movaps":
                return False, None

        spilled_regs = []
        allowed_spilled_regs = [
            capstone.x86.X86_REG_RDI,
            capstone.x86.X86_REG_RSI,
            capstone.x86.X86_REG_RDX,
            capstone.x86.X86_REG_RCX,
            capstone.x86.X86_REG_R8,
            capstone.x86.X86_REG_R9,
        ]
        for insn in reversed(self.project.factory.block(head.addr, size=head.size).capstone.insns[:-2]):
            if (
                insn.mnemonic == "mov"
                and insn.operands[0].type == capstone.x86.X86_OP_MEM
                and insn.operands[1].type == capstone.x86.X86_OP_REG
            ):
                spilled_regs.append(insn.operands[1].reg)
            else:
                break

        if not set(spilled_regs).issubset(set(allowed_spilled_regs)):
            return False, None

        i = next(
            (i for i, reg in enumerate(allowed_spilled_regs) if reg in spilled_regs),
            len(allowed_spilled_regs),
        )

        return True, i


register_analysis(CallingConventionAnalysis, "CallingConvention")



from __future__ import annotations
from angr.ailment import Expr, Stmt

from angr.calling_conventions import SimCCUsercall
from angr.engines.vex.claripy.ccall import data
from .rewriter_base import CCallRewriterBase


AMD64_CondTypes = data["AMD64"]["CondTypes"]
AMD64_OpTypes = data["AMD64"]["OpTypes"]
AMD64_CondBitMasks = data["AMD64"]["CondBitMasks"]
AMD64_CondBitOffsets = data["AMD64"]["CondBitOffsets"]


class AMD64CCallRewriter(CCallRewriterBase):
    """
    Implements VEX ccall rewriter for AMD64.
    """

    __slots__ = ()

    def _rewrite(self, ccall: Expr.VEXCCallExpression) -> Expr.Expression | None:
        if ccall.callee == "amd64g_calculate_condition":
            cond = ccall.operands[0]
            op = ccall.operands[1]
            dep_1 = ccall.operands[2]
            dep_2 = ccall.operands[3]
            if isinstance(cond, Expr.Const) and isinstance(op, Expr.Const):
                cond_v = cond.value_int
                op_v = op.value_int
                if cond_v == AMD64_CondTypes["CondLE"]:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 <=s dep_2
                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(ccall.idx, "CmpLE", (dep_1, dep_2), True, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if (
                        op_v
                        in {
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            AMD64_OpTypes["G_CC_OP_LOGICQ"],
                        }
                        and isinstance(dep_2, Expr.Const)
                        and dep_2.value == 0
                    ):
                        # dep_1 >=s 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(
                            ccall.idx,
                            "CmpLE",
                            (dep_1, dep_2),
                            True,
                            **ccall.tags,
                        )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif cond_v == AMD64_CondTypes["CondNLE"]:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 >s dep_2
                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(ccall.idx, "CmpGT", (dep_1, dep_2), True, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if (
                        op_v
                        in {
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            AMD64_OpTypes["G_CC_OP_LOGICQ"],
                        }
                        and isinstance(dep_2, Expr.Const)
                        and dep_2.value == 0
                    ):
                        # dep_1 >s 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(
                            ccall.idx,
                            "CmpGT",
                            (dep_1, dep_2),
                            True,
                            **ccall.tags,
                        )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif cond_v in {AMD64_CondTypes["CondZ"], AMD64_CondTypes["CondNZ"]}:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 - dep_2 == 0 or dep_1 - dep_2 != 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        expr_op = "CmpEQ" if cond_v == AMD64_CondTypes["CondZ"] else "CmpNE"

                        r = Expr.BinaryOp(ccall.idx, expr_op, (dep_1, dep_2), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        AMD64_OpTypes["G_CC_OP_LOGICQ"],
                    }:
                        # dep_1 == 0 or dep_1 != 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )
                        expr_op = "CmpEQ" if cond_v == AMD64_CondTypes["CondZ"] else "CmpNE"

                        r = Expr.BinaryOp(
                            ccall.idx, expr_op, (dep_1, Expr.Const(None, None, 0, dep_1.bits)), False, **ccall.tags
                        )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SHRB"],
                        AMD64_OpTypes["G_CC_OP_SHRW"],
                        AMD64_OpTypes["G_CC_OP_SHRL"],
                        AMD64_OpTypes["G_CC_OP_SHRQ"],
                    }:
                        # dep_1 == 0 or dep_1 != 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SHRB"],
                            AMD64_OpTypes["G_CC_OP_SHRW"],
                            AMD64_OpTypes["G_CC_OP_SHRL"],
                            ccall.tags,
                        )
                        expr_op = "CmpEQ" if cond_v == AMD64_CondTypes["CondZ"] else "CmpNE"

                        zero = Expr.Const(None, None, 0, dep_1.bits)
                        r = Expr.BinaryOp(ccall.idx, expr_op, (dep_1, zero), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if op_v == AMD64_OpTypes["G_CC_OP_COPY"]:
                        # dep_1 & G_CC_MASK_Z == 0 or dep_1 & G_CC_MASK_Z != 0

                        bitmask = AMD64_CondBitMasks["G_CC_MASK_Z"]
                        assert isinstance(bitmask, int)
                        flag = Expr.Const(None, None, bitmask, dep_1.bits)
                        masked_dep = Expr.BinaryOp(None, "And", [dep_1, flag], False, **ccall.tags)
                        zero = Expr.Const(None, None, 0, dep_1.bits)
                        expr_op = "CmpEQ" if cond_v == AMD64_CondTypes["CondZ"] else "CmpNE"

                        r = Expr.BinaryOp(ccall.idx, expr_op, (masked_dep, zero), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_DECB"],
                        AMD64_OpTypes["G_CC_OP_DECW"],
                        AMD64_OpTypes["G_CC_OP_DECL"],
                        AMD64_OpTypes["G_CC_OP_DECQ"],
                    }:
                        # dep_1 == 0 or dep_1 != 0
                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SHRB"],
                            AMD64_OpTypes["G_CC_OP_SHRW"],
                            AMD64_OpTypes["G_CC_OP_SHRL"],
                            ccall.tags,
                        )
                        expr_op = "CmpEQ" if cond_v == AMD64_CondTypes["CondZ"] else "CmpNE"

                        zero = Expr.Const(None, None, 0, dep_1.bits)
                        r = Expr.BinaryOp(ccall.idx, expr_op, (dep_1, zero), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif cond_v == AMD64_CondTypes["CondL"]:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 - dep_2 <s 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(ccall.idx, "CmpLT", (dep_1, dep_2), True, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        AMD64_OpTypes["G_CC_OP_LOGICQ"],
                    }:
                        # dep_1 is the result, dep_2 is always zero
                        # dep_1 <s 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_LOGICB"],
                            AMD64_OpTypes["G_CC_OP_LOGICW"],
                            AMD64_OpTypes["G_CC_OP_LOGICL"],
                            ccall.tags,
                        )
                        zero = Expr.Const(None, None, 0, dep_1.bits)
                        r = Expr.BinaryOp(ccall.idx, "CmpLT", (dep_1, zero), True, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

                elif cond_v == AMD64_CondTypes["CondNBE"]:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 - dep_2 > 0

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(ccall.idx, "CmpGT", (dep_1, dep_2), False, **ccall.tags)
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif cond_v == AMD64_CondTypes["CondB"]:
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_ADDB"],
                        AMD64_OpTypes["G_CC_OP_ADDW"],
                        AMD64_OpTypes["G_CC_OP_ADDL"],
                        AMD64_OpTypes["G_CC_OP_ADDQ"],
                    }:
                        # __CFADD__(dep_1, dep_2)

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_ADDB"],
                            AMD64_OpTypes["G_CC_OP_ADDW"],
                            AMD64_OpTypes["G_CC_OP_ADDL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_ADDB"],
                            AMD64_OpTypes["G_CC_OP_ADDW"],
                            AMD64_OpTypes["G_CC_OP_ADDL"],
                            ccall.tags,
                        )

                        return Stmt.Call(
                            ccall.idx,
                            "__CFADD__",
                            calling_convention=SimCCUsercall(self.project.arch, [], None),
                            args=[dep_1, dep_2],
                            bits=ccall.bits,
                            **ccall.tags,
                        )
                    if op_v in {
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        AMD64_OpTypes["G_CC_OP_SUBQ"],
                    }:
                        # dep_1 <u dep_2

                        dep_1 = self._fix_size(
                            dep_1,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )
                        dep_2 = self._fix_size(
                            dep_2,
                            op_v,
                            AMD64_OpTypes["G_CC_OP_SUBB"],
                            AMD64_OpTypes["G_CC_OP_SUBW"],
                            AMD64_OpTypes["G_CC_OP_SUBL"],
                            ccall.tags,
                        )

                        r = Expr.BinaryOp(
                            ccall.idx,
                            "CmpLT",
                            (dep_1, dep_2),
                            False,
                            **ccall.tags,
                        )
                        return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)
                elif (
                    cond_v == AMD64_CondTypes["CondS"]
                    and op_v
                    in {
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        AMD64_OpTypes["G_CC_OP_LOGICQ"],
                    }
                    and isinstance(dep_2, Expr.Const)
                    and dep_2.value == 0
                ):
                    # dep_1 < 0

                    dep_1 = self._fix_size(
                        dep_1,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        ccall.tags,
                    )
                    dep_2 = self._fix_size(
                        dep_2,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        ccall.tags,
                    )

                    r = Expr.BinaryOp(
                        ccall.idx,
                        "CmpLT",
                        (dep_1, dep_2),
                        True,
                        **ccall.tags,
                    )
                    return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

                elif (
                    cond_v == AMD64_CondTypes["CondNS"]
                    and op_v
                    in {
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        AMD64_OpTypes["G_CC_OP_LOGICQ"],
                    }
                    and isinstance(dep_2, Expr.Const)
                    and dep_2.value == 0
                ):
                    # dep_1 >= 0
                    dep_1 = self._fix_size(
                        dep_1,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        ccall.tags,
                    )
                    dep_2 = self._fix_size(
                        dep_2,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_LOGICB"],
                        AMD64_OpTypes["G_CC_OP_LOGICW"],
                        AMD64_OpTypes["G_CC_OP_LOGICL"],
                        ccall.tags,
                    )

                    r = Expr.BinaryOp(
                        ccall.idx,
                        "CmpGE",
                        (dep_1, dep_2),
                        True,
                        **ccall.tags,
                    )
                    return Expr.Convert(None, r.bits, ccall.bits, False, r, **ccall.tags)

        elif ccall.callee == "amd64g_calculate_rflags_c":
            # calculate the carry flag
            op = ccall.operands[0]
            dep_1 = ccall.operands[1]
            dep_2 = ccall.operands[2]
            ndep = ccall.operands[3]
            if isinstance(op, Expr.Const):
                op_v = op.value_int
                if op_v in {
                    AMD64_OpTypes["G_CC_OP_ADDB"],
                    AMD64_OpTypes["G_CC_OP_ADDW"],
                    AMD64_OpTypes["G_CC_OP_ADDL"],
                    AMD64_OpTypes["G_CC_OP_ADDQ"],
                }:
                    # pc_actions_ADD

                    dep_1 = self._fix_size(
                        dep_1,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_ADDB"],
                        AMD64_OpTypes["G_CC_OP_ADDW"],
                        AMD64_OpTypes["G_CC_OP_ADDL"],
                        ccall.tags,
                    )
                    dep_2 = self._fix_size(
                        dep_2,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_ADDB"],
                        AMD64_OpTypes["G_CC_OP_ADDW"],
                        AMD64_OpTypes["G_CC_OP_ADDL"],
                        ccall.tags,
                    )

                    return Expr.ITE(
                        None,
                        Expr.BinaryOp(
                            None,
                            "CmpLE",
                            [
                                Expr.BinaryOp(None, "Add", [dep_1, dep_2], False),
                                dep_1,
                            ],
                            False,
                        ),
                        Expr.Const(None, None, 0, ccall.bits),
                        Expr.Const(None, None, 1, ccall.bits),
                        **ccall.tags,
                    )

                if op_v in {
                    AMD64_OpTypes["G_CC_OP_SUBB"],
                    AMD64_OpTypes["G_CC_OP_SUBW"],
                    AMD64_OpTypes["G_CC_OP_SUBL"],
                    AMD64_OpTypes["G_CC_OP_SUBQ"],
                }:
                    # pc_actions_SUB

                    dep_1 = self._fix_size(
                        dep_1,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        ccall.tags,
                    )
                    dep_2 = self._fix_size(
                        dep_2,
                        op_v,
                        AMD64_OpTypes["G_CC_OP_SUBB"],
                        AMD64_OpTypes["G_CC_OP_SUBW"],
                        AMD64_OpTypes["G_CC_OP_SUBL"],
                        ccall.tags,
                    )

                    cf = Expr.BinaryOp(
                        None,
                        "CmpLT",
                        [
                            dep_1,
                            dep_2,
                        ],
                        False,
                    )
                    if cf.bits == ccall.bits:
                        return cf
                    return Expr.Convert(None, cf.bits, ccall.bits, False, cf, **ccall.tags)

                if op_v in {
                    AMD64_OpTypes["G_CC_OP_DECB"],
                    AMD64_OpTypes["G_CC_OP_DECW"],
                    AMD64_OpTypes["G_CC_OP_DECL"],
                    AMD64_OpTypes["G_CC_OP_DECQ"],
                }:
                    # pc_actions_DEC
                    bitmask = AMD64_CondBitMasks["G_CC_MASK_C"]
                    bitmask_1 = AMD64_CondBitOffsets["G_CC_SHIFT_C"]
                    assert isinstance(bitmask, int) and isinstance(bitmask_1, int)
                    return Expr.BinaryOp(
                        None,
                        "Shr",
                        [
                            Expr.BinaryOp(
                                None,
                                "And",
                                [ndep, Expr.Const(None, None, bitmask, 64)],
                                False,
                            ),
                            Expr.Const(None, None, bitmask_1, 64),
                        ],
                        False,
                        **ccall.tags,
                    )

        return None

    @staticmethod
    def _fix_size(expr, op_v: int, type_8bit, type_16bit, type_32bit, tags):
        if op_v == type_8bit:
            bits = 8
        elif op_v == type_16bit:
            bits = 16
        elif op_v == type_32bit:
            bits = 32
        else:
            bits = 64
        if bits < 64:
            if isinstance(expr, Expr.Const):
                return Expr.Const(expr.idx, None, expr.value_int & ((1 << bits) - 1), bits, **tags)
            return Expr.Convert(None, 64, bits, False, expr, **tags)
        return expr



from __future__ import annotations
import typing

from angr.ailment import AILBlockViewer

if typing.TYPE_CHECKING:
    from angr.ailment.expression import BinaryOp
    from angr.ailment.statement import Statement
    from angr.ailment.block import Block


class BooleanCounter(AILBlockViewer):
    """
    This class counts the number of Boolean operators an expression has.
    In the case of: `if (a || (b && c))`, it will count 2 Boolean operators.
    """

    def __init__(self):
        super().__init__()
        self.boolean_cnt = 0

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
        if expr.op in {"LogicalAnd", "LogicalOr"}:
            self.boolean_cnt += 1

        self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)



from __future__ import annotations

from enum import Enum

from .plugin import KnowledgeBasePlugin
from angr.sim_type import SimTypeFunction
from angr.calling_conventions import SimCC


class CallsitePrototypeKind(Enum):
    """
    Describes the type of a callsite prototype.
    """

    INFERRED = 1
    PROPAGATED = 2
    MANUAL = 3  # user specified


class CallsitePrototypes(KnowledgeBasePlugin):
    """
    CallsitePrototypes manages callee prototypes at call sites.
    """

    def __init__(self, kb):
        super().__init__(kb=kb)

        self._prototypes: dict[int, dict[CallsitePrototypeKind, tuple[SimCC, SimTypeFunction]]] = {}

    def set_prototype(
        self,
        callsite_block_addr: int,
        cc: SimCC,
        prototype: SimTypeFunction,
        *,
        kind: CallsitePrototypeKind = CallsitePrototypeKind.INFERRED,
        manual: bool = False,
        propagated: bool = False,
    ) -> None:
        if manual:
            kind = CallsitePrototypeKind.MANUAL
        elif propagated:
            kind = CallsitePrototypeKind.PROPAGATED

        if callsite_block_addr not in self._prototypes:
            self._prototypes[callsite_block_addr] = {}
        self._prototypes[callsite_block_addr][kind] = cc, prototype

    def get_cc(
        self, callsite_block_addr: int, *, kind: CallsitePrototypeKind = CallsitePrototypeKind.INFERRED
    ) -> SimCC | None:
        try:
            return self._prototypes[callsite_block_addr][kind][0]
        except KeyError:
            return None

    def get_prototype(
        self, callsite_block_addr: int, *, kind: CallsitePrototypeKind = CallsitePrototypeKind.INFERRED
    ) -> SimTypeFunction | None:
        try:
            return self._prototypes[callsite_block_addr][kind][1]
        except KeyError:
            return None

    def is_prototype_manual(self, callsite_block_addr: int) -> bool | None:
        try:
            return CallsitePrototypeKind.MANUAL in self._prototypes[callsite_block_addr]
        except KeyError:
            return None

    def is_prototype_certain(self, callsite_block_addr: int) -> bool | None:
        try:
            return (
                CallsitePrototypeKind.MANUAL in self._prototypes[callsite_block_addr]
                or CallsitePrototypeKind.PROPAGATED in self._prototypes[callsite_block_addr]
            )
        except KeyError:
            return None

    def has_some_prototype(self, callsite_block_addr: int) -> bool:
        return callsite_block_addr in self._prototypes and len(self._prototypes[callsite_block_addr]) > 0

    def has_prototype(
        self, callsite_block_addr: int, *, kind: CallsitePrototypeKind = CallsitePrototypeKind.INFERRED
    ) -> bool:
        return callsite_block_addr in self._prototypes and kind in self._prototypes[callsite_block_addr]

    def copy(self):
        o = CallsitePrototypes(self._kb)
        for callsite_addr, kinds in self._prototypes.items():
            o._prototypes[callsite_addr] = kinds.copy()
        return o


KnowledgeBasePlugin.register_default("callsite_prototypes", CallsitePrototypes)



from __future__ import annotations
import angr


class InitializeCriticalSectionAndSpinCount(angr.SimProcedure):
    def run(self, lpCriticalSection, dwSpinCount):
        return 1


class InitializeCriticalSectionEx(angr.SimProcedure):
    def run(self, lpCriticalSection, dwSpinCount, Flags):
        return 1



from __future__ import annotations
from .preset import DecompilationPreset
from angr.analyses.decompiler.optimization_passes import (
    RegisterSaveAreaSimplifier,
    StackCanarySimplifier,
    WinStackCanarySimplifier,
    BasePointerSaveSimplifier,
    ConstantDereferencesSimplifier,
    RetAddrSaveSimplifier,
    X86GccGetPcSimplifier,
    ITERegionConverter,
    ITEExprConverter,
    ExprOpSwapper,
    ReturnDuplicatorHigh,
    SwitchDefaultCaseDuplicator,
    LoweredSwitchSimplifier,
    ReturnDuplicatorLow,
    ReturnDeduplicator,
    FlipBooleanCmp,
    InlinedStringTransformationSimplifier,
    InlinedStrlenSimplifier,
    CallStatementRewriter,
    DeadblockRemover,
    SwitchReusedEntryRewriter,
    ConditionConstantPropagation,
    RegisterSaveAreaSimplifierAdvanced,
    DetermineLoadSizes,
    PostStructuringPeepholeOptimizationPass,
    EagerStdStringConcatenationPass,
    EagerStdStringEvalPass,
    StaticVVarRewriter,
)


preset_malware = DecompilationPreset(
    "malware",
    [
        RegisterSaveAreaSimplifier,
        StackCanarySimplifier,
        WinStackCanarySimplifier,
        RegisterSaveAreaSimplifierAdvanced,
        BasePointerSaveSimplifier,
        ConstantDereferencesSimplifier,
        RetAddrSaveSimplifier,
        X86GccGetPcSimplifier,
        ITERegionConverter,
        ITEExprConverter,
        ExprOpSwapper,
        ReturnDuplicatorHigh,
        DeadblockRemover,
        SwitchDefaultCaseDuplicator,
        SwitchReusedEntryRewriter,
        LoweredSwitchSimplifier,
        ReturnDuplicatorLow,
        ReturnDeduplicator,
        FlipBooleanCmp,
        InlinedStringTransformationSimplifier,
        InlinedStrlenSimplifier,
        CallStatementRewriter,
        ConditionConstantPropagation,
        DetermineLoadSizes,
        PostStructuringPeepholeOptimizationPass,
        EagerStdStringConcatenationPass,
        EagerStdStringEvalPass,
        StaticVVarRewriter,
    ],
)


__all__ = ["preset_malware"]



from __future__ import annotations
import logging

from angr.ailment import Statement, Block, Assignment, BinaryOp
from angr.ailment.expression import Const, VirtualVariable, Load
from angr.ailment.block_walker import AILBlockViewer, AILBlockRewriter
from angr.ailment.statement import Call
from angr.sim_type import SimTypeWideChar, SimTypeChar, SimTypePointer
from angr.utils.graph import GraphUtils
from .optimization_pass import OptimizationPass, OptimizationPassStage


_l = logging.getLogger(__name__)


#
# Data types
#


class FixedBuffer:
    def __init__(self, ident: str | None, size: int, content: bytes):
        self.ident = ident or "<unnamed>"
        self.size = size
        self.content = content

    def __repr__(self):
        return f"<FixedBuffer {self.ident} size={self.size}>"


class FixedBufferPtr:
    def __init__(self, buffer_ident: str, offset: int = 0):
        self.buffer_ident = buffer_ident
        self.offset = offset

    def __repr__(self):
        return f"<FixedBufferPtr {self.buffer_ident} + {self.offset}>"


class Offset:
    def __init__(self, value: int, bits: int):
        self.value = value
        self.bits = bits

    def __repr__(self):
        return f"<Offset {self.value} ({self.bits} bits)>"


#
# Visitors
#


class VVarRewritingVisitor(AILBlockRewriter):
    """
    The visitor that rewrites vvars and their reads.
    """

    def __init__(self, static_buffers: dict[str, FixedBuffer], static_vvars: dict[int, FixedBufferPtr | Const], kb):
        super().__init__(update_block=False)
        self._static_buffers = static_buffers
        self._static_vvars = static_vvars
        self.kb = kb

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement, block: Block | None
    ):
        if expr.varid in self._static_vvars:
            v = self._static_vvars[expr.varid]
            if isinstance(v, Const):
                return v
        return None

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Block | None):
        if isinstance(expr.addr, VirtualVariable) and expr.addr.varid in self._static_vvars:
            v = self._static_vvars[expr.addr.varid]
            if isinstance(v, FixedBufferPtr):
                buffer = self._static_buffers.get(v.buffer_ident, None)
                if buffer is None:
                    _l.warning("Cannot find static buffer %s", v.buffer_ident)
                    return None
                if v.offset < 0 or v.offset + expr.size > buffer.size:
                    _l.warning(
                        "Static buffer %s access out of bounds: offset=%d, size=%d, buffer_size=%d",
                        v.buffer_ident,
                        v.offset,
                        expr.size,
                        buffer.size,
                    )
                    return None
                data = buffer.content[v.offset : v.offset + expr.size]
                value = int.from_bytes(data, byteorder="little" if expr.endness == "Iend_LE" else "big")
                return Const(None, None, value, expr.bits, **expr.tags)

        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Block | None):
        if expr.target in {"strlen", "wcslen"} and expr.args:
            arg = expr.args[0]
            if isinstance(arg, VirtualVariable) and arg.varid in self._static_vvars:
                v = self._static_vvars[arg.varid]
                if isinstance(v, FixedBufferPtr):
                    buffer = self._static_buffers.get(v.buffer_ident, None)
                    if buffer is None:
                        _l.warning("Cannot find static buffer %s", v.buffer_ident)
                        return None
                    # compute the length of the string in the buffer
                    data = buffer.content[v.offset :]
                    str_len = 0
                    word_size = 2 if expr.target == "wcslen" else 1
                    while (
                        str_len < len(data)
                        and data[str_len * word_size : str_len * word_size + word_size] != b"\x00" * word_size
                    ):
                        str_len += 1
                    return Const(None, None, str_len, expr.bits, **expr.tags)

        elif expr.args and expr.prototype is not None:
            new_args: list | None = None
            for arg_idx, (arg_type, arg) in enumerate(zip(expr.prototype.args, expr.args)):
                if (  # noqa:SIM102
                    isinstance(arg, VirtualVariable)
                    and arg.varid in self._static_vvars
                    and isinstance(self._static_vvars[arg.varid], FixedBufferPtr)
                ):
                    if isinstance(arg_type, SimTypePointer):
                        if isinstance(arg_type.pts_to, SimTypeChar):
                            # char*
                            data = self._static_buffers[self._static_vvars[arg.varid].buffer_ident].content
                            offset = self._static_vvars[arg.varid].offset
                            str_bytes = bytearray()
                            idx = offset
                            while idx < len(data):
                                byte = data[idx : idx + 1]
                                if byte == b"\x00":
                                    break
                                str_bytes += byte
                                idx += 1
                            str_id = self.kb.custom_strings.allocate(bytes(str_bytes))
                            str_id_arg = Const(
                                None,
                                None,
                                str_id,
                                arg.bits,
                                custom_string=True,
                                **arg.tags,
                            )
                            if new_args is None:
                                new_args = expr.args[:arg_idx]
                            new_args.append(str_id_arg)
                            continue

                        if isinstance(arg_type.pts_to, SimTypeWideChar):
                            # wchar*
                            data = self._static_buffers[self._static_vvars[arg.varid].buffer_ident].content
                            offset = self._static_vvars[arg.varid].offset
                            str_bytes = bytearray()
                            idx = offset
                            while idx + 2 <= len(data):
                                wchar_bytes = data[idx : idx + 2]
                                if wchar_bytes == b"\x00\x00":
                                    break
                                str_bytes += wchar_bytes
                                idx += 2
                            str_id = self.kb.custom_strings.allocate(bytes(str_bytes))
                            str_id_arg = Const(
                                None,
                                None,
                                str_id,
                                arg.bits,
                                custom_string=True,
                                **arg.tags,
                            )
                            if new_args is None:
                                new_args = expr.args[:arg_idx]
                            new_args.append(str_id_arg)
                            continue

                if new_args is not None:
                    new_args.append(arg)

            if new_args is not None:
                return Call(
                    expr.idx,
                    expr.target,
                    calling_convention=expr.calling_convention,
                    prototype=expr.prototype,
                    args=new_args,
                    ret_expr=expr.ret_expr,
                    fp_ret_expr=expr.fp_ret_expr,
                    bits=expr.bits,
                    **expr.tags,
                )

        return None


class VVarAliasVisitor(AILBlockViewer):
    """
    The visitor that discovers const assignments and aliases of existing static vvars.
    """

    def __init__(self, static_buffers: dict[str, FixedBuffer], static_vvars: dict[int, FixedBufferPtr | Const], kb):
        super().__init__()
        self._static_buffers = static_buffers
        self._static_vvars = static_vvars
        self.kb = kb

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
        src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        dst = stmt.dst
        if (
            isinstance(dst, VirtualVariable)
            and dst.varid not in self._static_vvars
            and src is not None
            and isinstance(src, FixedBufferPtr)
        ):
            self._static_vvars[dst.varid] = src

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement, block: Block | None
    ):
        return self._static_vvars.get(expr.varid, None)

    def _handle_Const(self, expr_idx: int, expr: Const, stmt_idx: int, stmt: Statement, block: Block | None):
        return Offset(expr.value, expr.bits)

    def _handle_Call(self, stmt_idx: int, stmt: Call, block: Block | None):
        if (
            stmt.target == "memcpy"
            and isinstance(stmt.args[0], VirtualVariable)
            and isinstance(stmt.args[1], Const)
            and isinstance(stmt.args[2], Const)
        ):
            # got a new memcpy call that we can handle
            dst, src, size = stmt.args
            if dst.varid not in self._static_vvars:
                if src.tags.get("custom_string", False):
                    ident = f"static_buf_{stmt.tags['ins_addr']}"
                    buf = self.kb.custom_strings[src.value_int]
                    fixed_buffer = FixedBuffer(ident, size.value_int, buf)
                else:
                    # TODO: Support other cases
                    return super()._handle_Call(stmt_idx, stmt, block)
                if ident not in self._static_buffers:
                    self._static_buffers[ident] = fixed_buffer
                self._static_vvars[dst.varid] = FixedBufferPtr(ident, 0)

        return super()._handle_Call(stmt_idx, stmt, block)

    def _handle_BinaryOp(self, expr_idx: int, expr: BinaryOp, stmt_idx: int, stmt: Statement, block: Block | None):
        op0 = self._handle_expr(0, expr.operands[0], stmt_idx, stmt, block)
        op1 = self._handle_expr(1, expr.operands[1], stmt_idx, stmt, block)
        if op0 is None or op1 is None:
            return None
        if expr.op == "Add":
            if isinstance(op0, FixedBufferPtr) and isinstance(op1, Offset):
                return FixedBufferPtr(op0.buffer_ident, op0.offset + op1.value)
            if isinstance(op1, FixedBufferPtr) and isinstance(op0, Offset):
                return FixedBufferPtr(op1.buffer_ident, op1.offset + op0.value)
        elif expr.op == "Sub":
            if isinstance(op0, FixedBufferPtr) and isinstance(op1, Offset):
                return FixedBufferPtr(op0.buffer_ident, op0.offset - op1.value)
        elif expr.op == "Mul":
            if isinstance(op0, Offset) and isinstance(op1, Offset):
                return Offset(op0.value * op1.value, expr.bits)
        return None


#
# The main class
#


class StaticVVarRewriter(OptimizationPass):
    """
    Rewrite user-specified vvars as static values or fix-sized buffers. Also rewrites reads from pointers derived off
    of such vvars.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.BEFORE_VARIABLE_RECOVERY
    NAME = "Static virtual variable rewriter"
    DESCRIPTION = __doc__.strip()

    def __init__(
        self,
        func,
        static_buffers: dict[str, FixedBuffer] | None = None,
        static_vvars: dict[int, FixedBufferPtr | Const] | None = None,
        **kwargs,
    ):
        super().__init__(func, **kwargs)
        self._static_buffers = static_buffers
        self._static_vvars = static_vvars
        self.analyze()

    def _check(self):
        # discover aliases
        alias_visitor = VVarAliasVisitor(self._static_buffers, self._static_vvars, self.kb)
        head = {(node.addr, node.idx): node for node in self._graph}[self.entry_node_addr]
        for block in reversed(list(GraphUtils.dfs_postorder_nodes_deterministic(self._graph, head))):
            alias_visitor.walk(block)
        if not self._static_buffers or not self._static_vvars:
            return False, None
        return True, None

    def _analyze(self, cache=None):
        # rewrite vvars
        g = self._graph

        rewritten = False
        while True:
            rewriter = VVarRewritingVisitor(self._static_buffers, self._static_vvars, self.kb)
            for block in list(g):
                new_block = rewriter.walk(block)
                if new_block is not None:
                    rewritten = True
                    self._update_block(block, new_block)

            if not rewritten:
                break

            g = self.out_graph
            # discover more aliases
            old_static_buffers = dict(self._static_buffers)
            old_static_vvars = dict(self._static_vvars)
            alias_visitor = VVarAliasVisitor(self._static_buffers, self._static_vvars, self.kb)
            head = {(node.addr, node.idx): node for node in g}[self.entry_node_addr]
            for block in reversed(list(GraphUtils.dfs_postorder_nodes_deterministic(g, head))):
                alias_visitor.walk(block)

            if old_static_buffers == self._static_buffers and old_static_vvars == self._static_vvars:
                # no new aliases found
                break



from __future__ import annotations


class TraversalState:
    """
    The abstract state for the traversal engine.
    """

    def __init__(
        self,
        arch,
        func,
        live_registers: set[int] | None = None,
        live_stackvars: set[tuple[int, int]] | None = None,
    ):
        self.arch = arch
        self.func = func

        self.live_registers: set[int] = set() if live_registers is None else live_registers
        self.live_stackvars: set[tuple[int, int]] = set() if live_stackvars is None else live_stackvars
        self.live_tmps: set[int] = set()  # tmps are internal to a block only and never propagated from another state

    def copy(self) -> TraversalState:
        return TraversalState(
            self.arch,
            self.func,
            live_registers=self.live_registers.copy(),
            live_stackvars=self.live_stackvars.copy(),
        )

    def merge(self, *others: TraversalState) -> bool:
        merge_occurred = False

        all_regs: set[int] = self.live_registers.copy()
        for o in others:
            if o.live_registers.difference(all_regs):
                merge_occurred = True
            all_regs |= o.live_registers

        all_stackvars: set[tuple[int, int]] = self.live_stackvars.copy()
        for o in others:
            if o.live_stackvars.difference(all_stackvars):
                merge_occurred = True
            all_stackvars |= o.live_stackvars

        self.live_registers = all_regs
        self.live_stackvars = all_stackvars
        return merge_occurred



from __future__ import annotations


def decode_utf16_string(data: bytes) -> str:
    """
    Decode a UTF-16 encoded string from a bytes object in a resilient manner.

    :param data: The bytes object containing the UTF-16 encoded string.
    :param errors: The error handling scheme. Default is 'strict'.
                   Other options include 'ignore', 'replace', etc.
    :return: The decoded string.
    """
    if len(data) % 2 == 1:
        data = data[:-1]  # Trim off the last byte if the length is odd

    # If no BOM, try to decode as little-endian first
    try:
        return data.decode("utf-16-le")
    except UnicodeDecodeError:
        return "<utf16-decode-error>"



from __future__ import annotations
import angr

from cle.backends.externs.simdata.io_file import io_file_data_for_arch


class fwrite(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, src, size, nmemb, file_ptr):
        fd_offset = io_file_data_for_arch(self.state.arch)["fd"]
        fileno = self.state.mem[file_ptr + fd_offset :].int.resolved
        simfd = self.state.posix.get_fd(fileno)
        if simfd is None:
            return -1
        return simfd.write(src, size * nmemb)


fwrite_unlocked = fwrite



from __future__ import annotations
import claripy

from angr.storage.memory_mixins.memory_mixin import MemoryMixin


class PermissionsMixin(MemoryMixin):
    """
    This mixin adds a permissions_bits field and properties for extracting the read/write/exec permissions. It does NOT
    add permissions checking.
    """

    def __init__(self, permissions: int | claripy.ast.BV | None = None, **kwargs):
        super().__init__(**kwargs)
        if permissions is None:
            permissions = 7
        if isinstance(permissions, int):
            permissions = claripy.BVV(permissions, 3)
        self.permission_bits = permissions

    def copy(self, memo):
        o = super().copy(memo)
        o.permission_bits = self.permission_bits
        return o

    @property
    def perm_read(self):
        return self.permission_bits & 1

    @property
    def perm_write(self):
        return self.permission_bits & 2

    @property
    def perm_exec(self):
        return self.permission_bits & 4



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
import logging


from angr.ailment.statement import Assignment
from angr.ailment.expression import VirtualVariable
from angr.code_location import CodeLocation, ExternalCodeLocation
from angr.analyses.decompiler.stack_item import StackItem, StackItemType
from angr.utils.ail import is_phi_assignment
from .optimization_pass import OptimizationPass, OptimizationPassStage

_l = logging.getLogger(name=__name__)


class RegisterSaveAreaSimplifierAdvanced(OptimizationPass):
    """
    Optimizes away registers that are stored to or restored on the stack space.

    This analysis is more complex than RegisterSaveAreaSimplifier because it handles:
    (1) Registers that are stored in the stack shadow space (sp+N) according to the Windows x64 calling convention.
    (2) Registers that are aliases of sp.
    """

    ARCHES = None
    PLATFORMS = None
    STAGE = OptimizationPassStage.AFTER_MAKING_CALLSITES
    NAME = "Simplify register save areas (advanced)"
    DESCRIPTION = __doc__.strip()  # type:ignore

    def __init__(self, func, **kwargs):
        super().__init__(func, **kwargs)
        self._srda = None

        self.analyze()

    def _check(self):

        self._srda = self.project.analyses.SReachingDefinitions(
            subject=self._func, func_graph=self._graph, func_args=self._arg_vvars
        )
        info = self._find_reg_store_and_restore_locations()
        if not info:
            return False, None

        return True, {"info": info}

    @staticmethod
    def _modify_statement(
        old_block, stmt_idx_: int, updated_blocks_, stack_offset: int | None = None
    ):  # pylint:disable=unused-argument
        if old_block not in updated_blocks_:
            block = old_block.copy()
            updated_blocks_[old_block] = block
        else:
            block = updated_blocks_[old_block]
        block.statements[stmt_idx_] = None

    def _analyze(self, cache=None):

        if cache is None:
            return

        info: list[tuple[int, CodeLocation, int, CodeLocation, int]] = cache["info"]
        updated_blocks = {}

        for _regvar, regvar_loc, _stackvar, stackvar_loc, _ in info:
            # remove storing statements
            old_block = self._get_block(regvar_loc.block_addr, idx=regvar_loc.block_idx)
            assert regvar_loc.stmt_idx is not None
            self._modify_statement(old_block, regvar_loc.stmt_idx, updated_blocks)
            old_block = self._get_block(stackvar_loc.block_addr, idx=stackvar_loc.block_idx)
            assert stackvar_loc.stmt_idx is not None
            self._modify_statement(old_block, stackvar_loc.stmt_idx, updated_blocks)

        for old_block, new_block in updated_blocks.items():
            # remove all statements that are None
            new_block.statements = [stmt for stmt in new_block.statements if stmt is not None]
            # update it
            self._update_block(old_block, new_block)

        if updated_blocks:
            # update stack_items
            for _, _, _, _, stack_offset in info:
                self.stack_items[stack_offset] = StackItem(
                    stack_offset, self.project.arch.bytes, "regs", StackItemType.SAVED_REGS
                )

    def _find_reg_store_and_restore_locations(self) -> list[tuple[int, CodeLocation, int, CodeLocation, int]]:
        results = []

        assert self._srda is not None
        srda_model = self._srda.model
        # find all registers that are defined externally and used exactly once
        saved_vvars: set[tuple[int, CodeLocation]] = set()
        for vvar_id, loc in srda_model.all_vvar_definitions.items():
            if isinstance(loc, ExternalCodeLocation):
                uses = srda_model.all_vvar_uses.get(vvar_id, [])
                if len(uses) == 1:
                    vvar, used_loc = next(iter(uses))
                    if vvar is not None and vvar.was_reg:
                        saved_vvars.add((vvar_id, used_loc))

        if not saved_vvars:
            return results

        # for each candidate, we check to ensure:
        # - it is stored onto the stack (into a stack virtual variable)
        # - the stack virtual variable is only used once and restores the value to the same register
        # - the restore location is in the dominance frontier of the store location
        for vvar_id, used_loc in saved_vvars:
            def_block = self._get_block(used_loc.block_addr, idx=used_loc.block_idx)
            assert def_block is not None and used_loc.stmt_idx is not None
            stmt = def_block.statements[used_loc.stmt_idx]
            if not (
                isinstance(stmt, Assignment)
                and isinstance(stmt.dst, VirtualVariable)
                and stmt.dst.was_stack
                and isinstance(stmt.src, VirtualVariable)
                and stmt.src.was_reg
                and stmt.src.varid == vvar_id
            ):
                continue
            stack_vvar = stmt.dst
            all_stack_vvar_uses = srda_model.all_vvar_uses.get(stack_vvar.varid, [])
            # eliminate the use location if it's a phi statement
            stack_vvar_uses = set()
            for vvar_, loc_ in all_stack_vvar_uses:
                use_block = self._get_block(loc_.block_addr, idx=loc_.block_idx)
                if use_block is None or loc_.stmt_idx is None:
                    continue
                use_stmt = use_block.statements[loc_.stmt_idx]
                if is_phi_assignment(use_stmt):
                    continue
                stack_vvar_uses.add((vvar_, loc_))
            if len(stack_vvar_uses) != 1:
                continue
            _, stack_vvar_use_loc = next(iter(stack_vvar_uses))
            restore_block = self._get_block(stack_vvar_use_loc.block_addr, idx=stack_vvar_use_loc.block_idx)
            assert restore_block is not None
            restore_stmt = restore_block.statements[stack_vvar_use_loc.stmt_idx]

            if not (
                isinstance(restore_stmt, Assignment)
                and isinstance(restore_stmt.src, VirtualVariable)
                and restore_stmt.src.varid == stack_vvar.varid
                and isinstance(restore_stmt.dst, VirtualVariable)
                and restore_stmt.dst.was_reg
                and restore_stmt.dst.reg_offset == stmt.src.reg_offset
            ):
                continue
            # this is the dumb version of the dominance frontier check
            if self._within_dominance_frontier(def_block, restore_block, True, True):
                results.append(
                    (stmt.src.varid, used_loc, stack_vvar.varid, stack_vvar_use_loc, stack_vvar.stack_offset)
                )

        return results

    def _within_dominance_frontier(self, dom_node, node, use_preds: bool, use_succs: bool) -> bool:
        if use_succs:
            # scan forward
            succs = [succ for succ in self._graph.successors(dom_node) if succ is not dom_node]
            if len(succs) == 1:
                succ = succs[0]
                succ_preds = [pred for pred in self._graph.predecessors(succ) if pred is not succ]
                if len(succ_preds) == 0:
                    # the successor has no other predecessors
                    r = self._within_dominance_frontier(succ, node, False, True)
                    if r:
                        return True

                else:
                    # the successor has other predecessors; gotta step back
                    preds = [pred for pred in self._graph.predecessors(node) if pred is not node]
                    if len(preds) == 1 and preds[0] is node:
                        return True
            elif len(succs) == 2:
                return any(succ is node for succ in succs)

        if use_preds:
            # scan backward
            preds = [pred for pred in self._graph.predecessors(dom_node) if pred is not dom_node]
            if len(preds) == 1:
                pred = preds[0]
                pred_succs = [succ for succ in self._graph.successors(pred) if succ is not pred]
                if len(pred_succs) == 0:
                    # the predecessor has no other successors
                    return self._within_dominance_frontier(pred, node, True, False)

                # the predecessor has other successors; gotta step forward
                succs = [succ for succ in self._graph.successors(node) if succ is not node]
                if len(succs) == 1:
                    return self._graph.has_edge(node, succs[0])
            elif len(preds) == 2:
                return False

        return False



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from angr.ailment.expression import Load, BinaryOp, VirtualVariable, Const

from .base import PeepholeOptimizationExprBase


class RewriteMipsGpLoads(PeepholeOptimizationExprBase):
    """
    Rewrite $gp-based loads to their actual values on MIPS.
    """

    __slots__ = ()

    NAME = "MIPS GP-based Loads Rewriter"
    expr_classes = (Load,)

    def optimize(self, expr: Load, **kwargs):
        # Load(addr=(gp<8> - 0x7fc0<64>), size=8, endness=Iend_LE) - replace it with gp for this function
        if self.project.arch.name not in {"MIPS32", "MIPS64"}:
            return None
        if "gp" not in self.project.kb.functions[self.func_addr].info:
            return None

        gp_offset = self.project.arch.registers["gp"][0]
        if (
            expr.size == self.project.arch.bytes
            and isinstance(expr.addr, BinaryOp)
            and expr.addr.op in {"Add", "Sub"}
            and len(expr.addr.operands) == 2
            and isinstance(expr.addr.operands[0], VirtualVariable)
            and expr.addr.operands[0].was_reg
            and expr.addr.operands[0].reg_offset == gp_offset
            and isinstance(expr.addr.operands[1], Const)
        ):
            # just do the load...
            gp_value = self.project.kb.functions[self.func_addr].info["gp"]
            if expr.addr.op == "Add":
                addr = gp_value + expr.addr.operands[1].value
            else:  # Sub
                addr = gp_value - expr.addr.operands[1].value
            if self.project.arch.bits == 32:
                addr &= 0xFFFF_FFFF
            else:
                addr &= 0xFFFF_FFFF_FFFF_FFFF
            value = self.project.loader.memory.unpack_word(addr, size=expr.size)
            return Const(None, None, value, expr.size * self.project.arch.byte_width, **expr.tags)

        return None



from __future__ import annotations
from .base import SimSootExpr


class SimSootExpr_StaticFieldRef(SimSootExpr):
    def _execute(self):
        field_ref = self._translate_value(self.expr)
        self.expr = self.state.memory.load(field_ref, none_if_missing=True)



from __future__ import annotations
import logging
from collections import defaultdict

import networkx as nx

from angr.ailment.block import Block
from angr.ailment.statement import ConditionalJump

from .errors import SAILRSemanticError
from .similarity import ail_similarity_to_orig_blocks
from .utils import (
    copy_graph_and_nodes,
    replace_node_in_graph,
    ail_block_from_stmts,
    correct_jump_targets,
    deepcopy_ail_anyjump,
)

_l = logging.getLogger(name=__name__)


class AILBlockSplit:
    """
    This class represents a block that has been split into three parts, which is best explained in the
    AILMergeGraph class example. See that class for more information.

    The up_split, is all statements above the matched Longest Common Sequence (LCS), the match_split is the LCS,
    and the down_split is all statements below the matched LCS. This class should only be used in the context of
    the AILMergeGraph class.
    """

    def __init__(self, original=None, up_split=None, match_split=None, down_split=None):
        """

        :param original:
        :param up_split:   The block split above the matched LCS
        :param match_split:   The block containing only the matched LCS
        :param down_split:  The block split below the matched LCS
        """
        self.original = original
        self.up_split = up_split
        self.match_split = match_split
        self.down_split = down_split

    @classmethod
    def from_block_lcs(cls, original_block: Block, idx, len_):
        pre, mid, post = cls.split_ail_block(original_block, idx, len_)
        return cls(
            original=original_block,
            up_split=pre,
            match_split=mid,
            down_split=post,
        )

    @staticmethod
    def split_ail_block(block, split_idx, split_len) -> tuple[Block | None, Block | None, Block | None]:
        if split_len == len(block.statements):
            return None, block, None

        up_split = ail_block_from_stmts(block.statements[:split_idx], block_addr=block.addr)
        match_split = ail_block_from_stmts(
            block.statements[split_idx : split_idx + split_len], block_addr=None if up_split else block.addr
        )
        down_split = ail_block_from_stmts(block.statements[split_idx + split_len :])

        return up_split, match_split, down_split

    def __str__(self):
        return (
            f"<AILBlockSplit: OG: {self.original.__repr__()} | Up: {self.up_split.__repr__()} | "
            f"Match: {self.match_split.__repr__()} | Down: {self.down_split.__repr__()}>"
        )

    def __repr__(self):
        return self.__str__()


class AILMergeGraph:
    """
    This class represents the merged results of two AIL graphs that have been found to be similar. We can reference
    these graphs as G1 and G2. The two graphs are of the following form, where any node other than D can be empty:

        A
       / \
      B   C
      \\  /
       D
      / \
     E   F

    The D node can be a subgraph, but in both G1 and G2, this D-subgraph are exact duplicates of each
    other, except their top and bottom statements. This class is the result of merging those two D subgraphs.

    To explain that last part about statements differing at the ends, see this example:

    D1:
    -----
    a = 10;
    puts(a);
    puts("bye");
    -----

    D2:
    -----
    a = 11;
    puts(a);
    puts("cya");
    -----

    In this case, the merged D would contain just `puts(a)`. The statements above it, referred to as up_split in
    the code, and the statements below it, referred to as down_split in the code, would be moved out of the block
    and bounded by the conditions that lead to those statements. This creates a graph even in the case of a single
    block being the original D.

    Lastly, since this class will deal a lot with splitting blocks into pieces, we keep a mapping of how the original
    blocks turned into the new ones and vice versa.
    """

    def __init__(
        self, graph=None, original_graph=None, conditional_block=None, original_blocks=None, original_split_blocks=None
    ):
        self.graph = graph or nx.DiGraph()
        self.original_graph = original_graph or nx.DiGraph()
        self.conditional_block = conditional_block

        # All the original blocks in both graphs that are going to be merged into a single
        # graph that is self.graph
        self.original_blocks: dict[Block, list[Block]] = original_blocks or {}

        # The mapping from an original_block in original_blocks to an AILBlockSplit,
        # which can have an up, match, and down split. This happens when only partial
        # blocks match from the original blocks
        self.original_split_blocks: dict[Block, list[AILBlockSplit]] = original_split_blocks or {}

        # The mapping from each block in the self.graph to the original blocks that
        # are contained in the larger CFG
        self.merge_blocks_to_originals: dict[Block, set[Block | AILBlockSplit]] = defaultdict(set)
        self.merge_end_to_start: dict[Block, Block] = {}
        self.starts = []
        self.original_ends = []

    def create_conditionless_graph(self, starting_blocks: list[Block], graph_lcs) -> dict[Block, Block] | None:
        # get all the original blocks (reverted from the LCS) and their split blocks.
        # split-blocks are blocks that need to be split at some stmt index to make the two blocks
        # equal across both graphs. At a highlevel, the first block in both matching graphs either need
        # to be a full match or a spilt block with a None up-split (since the up-split represent a starting
        # stmt that mismatches).
        self.starts = starting_blocks.copy()
        merge_base, other_base = self.starts[:]
        for block in self.starts:
            og_blks_bfs_order, split_blks = ail_similarity_to_orig_blocks(block, graph_lcs, self.original_graph)
            self.original_blocks[block] = og_blks_bfs_order
            self.original_split_blocks[block] = [
                AILBlockSplit.from_block_lcs(og, idx, len(lcs)) for og, (lcs, idx) in split_blks.items()
            ]

        # eliminate shared blocks that are the same in original blocks
        shared_blocks = set(self.original_blocks[merge_base]).intersection(set(self.original_blocks[other_base]))
        for shared_block in shared_blocks:
            for block in self.starts:
                if shared_block not in self.original_blocks[block]:
                    continue

                self.original_blocks[block].remove(shared_block)
                if shared_block in self.original_split_blocks[block]:
                    self.original_split_blocks[block].remove(shared_block)

        # we now know all the original blocks that will be merged, in order of BFS
        # so let's create the graph that will be the final output merged graph
        #
        # we start by finding all the blocks that are about to be split, i.e., blocks that
        # have at least one stmt mismatching from each other
        base_to_split = {
            split_block.original: split_block.match_split
            for split_block in self.original_split_blocks[merge_base]
            if split_block.match_split is not None
        }

        # we create a new graph, full of the original blocks of the base, with blocks
        # that should be split replaced.
        # this graph is only the initial merge_graph needed, where only the blocks
        subgraph = nx.subgraph(self.original_graph, self.original_blocks[merge_base])
        # ensure all base blocks are within the subgraph
        for block in base_to_split:
            if block not in subgraph:
                return None
        self.graph, update_blocks = self.clone_graph_replace_splits(subgraph, base_to_split)
        self._update_all_split_refs(update_blocks)
        for update_block, new_block in update_blocks.items():
            if update_block in starting_blocks:
                idx = self.starts.index(update_block)
                self.starts[idx] = new_block
        merge_base, other_base = self.starts[:]

        # Add all the blocks that start the graph that may have been split
        # (i.e., only blocks in the blocks list for this function)
        for block in self.starts:
            for split_block in self.original_split_blocks[block]:
                if split_block.up_split:
                    merge_base_split = self._find_split_block_by_original(merge_base) or merge_base
                    merge_base_split = (
                        merge_base_split.match_split
                        if isinstance(merge_base_split, AILBlockSplit)
                        else merge_base_split
                    )

                    self.graph.add_edge(split_block.up_split, merge_base_split)

        return update_blocks

    def add_edges_to_condition(self, conditional_block, true_target, merge_end_pairs):
        for match_node, merge_end_pair in merge_end_pairs.items():
            cond_copy = conditional_block.copy()
            cond_copy.statements = [deepcopy_ail_anyjump(cond_copy.statements[0])]
            cond_copy.idx += 1

            # fix the condition edges
            cond_jump_stmt: ConditionalJump = cond_copy.statements[-1]
            if not merge_end_pair:
                raise SAILRSemanticError("Encountered a conditional jump that has no successors! This can be bad!")
            if len(merge_end_pair) == 1:
                b0 = merge_end_pair[0]
                b0_og = next(iter(self.merge_blocks_to_originals[b0]))
                if isinstance(b0_og, AILBlockSplit):
                    b0_og = b0_og.original

                b1_og = self._find_block_pair_in_originals(b0_og)
                if b1_og is None:
                    raise SAILRSemanticError(
                        f"Encountered a conditional jump that has only 1 successor on {self.starts}!"
                    )

                b1_og_succs = list(self.original_graph.successors(b1_og))
                if len(b1_og_succs) != 1:
                    raise SAILRSemanticError(
                        "Encountered a merge-end pair which ends in a return, this should be skipped!"
                    )

                b1 = b1_og_succs[0]
            else:
                b0, b1 = merge_end_pair

            if true_target == self._find_og_start_by_merge_end(b0):
                cond_jump_stmt.true_target.value = b0.addr
                cond_jump_stmt.false_target.value = b1.addr
            else:
                cond_jump_stmt.false_target.value = b0.addr
                cond_jump_stmt.true_target.value = b1.addr

            self.graph.add_edge(match_node, cond_copy)
            self.graph.add_edge(cond_copy, b0)
            self.graph.add_edge(cond_copy, b1)

    def create_mapping_to_merge_graph(self, updated_blocks: dict[Block, Block], start_blocks: list[Block]):
        merge_base, other_base = self.starts[:]

        # collect all the normal blocks, which should be easy to pair
        for base, other in zip(self.original_blocks[merge_base], self.original_blocks[other_base]):
            if base in updated_blocks:
                base = updated_blocks[base]
            self.merge_blocks_to_originals[base] = {base, other}

        # collect all the split blocks, which have new addrs, and match them to merge_base
        for sblock in self.original_split_blocks[merge_base]:
            attrs = ("up_split", "match_split", "down_split")
            for attr in attrs:
                sblock_split = getattr(sblock, attr)
                if not sblock_split:
                    continue
                self.merge_blocks_to_originals[sblock_split].add(sblock)

        # finally, use the info from merge_base mapping to add to the mapping for other_base
        for sblock in self.original_split_blocks[other_base]:
            attrs = ("up_split", "match_split", "down_split")
            base_original = self._find_merge_block_by_original(sblock.original)
            for attr in attrs:
                sblock_split = getattr(sblock, attr)
                if not sblock_split:
                    continue

                if attr != "match_split":
                    self.merge_blocks_to_originals[sblock_split].add(sblock)
                    continue

                for base_split, base_sblocks in self.merge_blocks_to_originals.items():
                    # this split block must have only one pair right now
                    if len(base_sblocks) > 1:
                        continue

                    base_sblock = next(iter(base_sblocks))
                    # the original base must match
                    if base_sblock.original != base_original:
                        continue

                    # this split block must also be the right split (up, match, down)
                    if base_split != getattr(base_sblock, attr):
                        continue

                    base_sblocks.add(sblock)

        # remove any extra block that will not be in the graph because it was split up
        deletable_blocks = set()
        end_pair_map = {}
        end_pairs = set()
        merge_to_end_pair = {}
        for sblocks in self.merge_blocks_to_originals.values():
            for sblock in sblocks:
                if isinstance(sblock, AILBlockSplit) and sblock.original in self.merge_blocks_to_originals:
                    deletable_blocks.add(sblock.original)
        for del_block in deletable_blocks:
            b0, b1 = self.merge_blocks_to_originals[del_block]
            end_pair_map[b0] = b1
            end_pair_map[b1] = b0
            end_pairs.add((b0, b1))
            merge_to_end_pair[del_block] = (b0, b1)
            del self.merge_blocks_to_originals[del_block]

        # mappings are collected, now we need to connect the merge ends of the graph to the upper parts with
        # the conditions that blocked the original
        merge_ends = {
            block: split for block, split in self.merge_blocks_to_originals.items() if block not in self.graph
        }
        assert all(len(bs) == 1 for block, bs in self.merge_blocks_to_originals.items() if block in merge_ends)
        self.merge_end_to_start = {merge_end: self._find_og_start_by_merge_end(merge_end) for merge_end in merge_ends}
        merge_end_pairs = defaultdict(list)
        for merge_end, split_blocks in merge_ends.items():
            split_block = next(iter(split_blocks))
            if split_block.original in self.original_blocks[merge_base]:
                merge_end_pairs[split_block.match_split].append(merge_end)
                continue

            merge_original = end_pair_map.get(split_block.original, None) or split_block.original
            base_split = self._find_split_block_by_original(merge_original)
            merge_end_pairs[base_split.match_split].append(merge_end)

        for pair in end_pairs:
            for block in pair:
                self.original_ends.append(block)
        # the case of single block merges
        if not self.original_ends:
            self.original_ends = start_blocks

        # moved here
        for _, pair in merge_to_end_pair.items():
            for block in pair:
                other_block = pair[0] if pair[1] is block else pair[1]
                while True:
                    for merge, originals in self.merge_blocks_to_originals.items():
                        if self._block_in_originals(merge, other_block):
                            continue

                        if len(originals) == 1:
                            og = next(iter(originals))
                            if isinstance(og, Block) and og == block:
                                originals.add(other_block)
                                break

                        found = False
                        for og in originals:
                            if isinstance(og, AILBlockSplit) and og.original == block and merge == og.match_split:
                                originals.add(other_block)
                                found = True
                                break

                        if found:
                            break

                    else:
                        break

        return merge_end_pairs

    def merged_is_split_type(self, merge_block: Block, split_type: str):
        if split_type not in ["up_split", "down_split", "match_split"]:
            raise ValueError("Must use a supported split type (up_split, down_split, match_split)")

        for oblock in self.merge_blocks_to_originals[merge_block]:
            if not isinstance(oblock, AILBlockSplit):
                continue

            if getattr(oblock, split_type) == merge_block:
                return True

        return False

    #
    # Private Functions
    #

    def _find_block_pair_in_originals(self, block: Block):
        for _, originals in self.merge_blocks_to_originals.items():
            # need at least 2 for a pair
            if len(originals) < 2:
                continue

            # the block we are searching for a pair for should exist in the originals
            for og in originals:
                if isinstance(og, Block) and og == block:
                    break
                if isinstance(og, AILBlockSplit) and og.original == block:
                    break
            else:
                return None

            # we now know that our target block is in this originals set
            # now we just need to find any other block that is not itself
            for other_block in originals:
                if isinstance(other_block, Block) and other_block != block:
                    return other_block
                if isinstance(other_block, AILBlockSplit) and other_block.original != block:
                    return other_block.original

        return None

    def _block_in_originals(self, merge_block: Block, target_block: Block):
        for original in self.merge_blocks_to_originals[merge_block]:
            if (isinstance(original, Block) and original == target_block) or (
                isinstance(original, AILBlockSplit) and original.original == target_block
            ):
                return True

        return False

    def _update_all_split_refs(self, update_map: dict[Block, Block]):
        for original, updated in update_map.items():
            for k in list(self.original_split_blocks.keys()):
                if k == original:
                    self.original_split_blocks[updated] = self.original_split_blocks[k]
                    del self.original_split_blocks[k]

            for v in self.original_split_blocks.values():
                for sblock in v:
                    for attr in ["up_split", "match_split", "down_split"]:
                        if getattr(sblock, attr) == original:
                            setattr(sblock, attr, updated)

            for k in list(self.original_blocks.keys()):
                if k == original:
                    self.original_blocks[updated] = self.original_blocks[k]
                    del self.original_blocks[k]

    def _find_merge_block_by_original(self, block: Block):
        for merge_block, originals in self.merge_blocks_to_originals.items():
            for og in originals:
                if isinstance(og, AILBlockSplit) and og.original == block:
                    return merge_block

                if isinstance(og, Block) and og == block:
                    return merge_block

        _l.warning("Error in finding the merge block from the original block on %s", block)
        return None

    def _find_split_block_by_original(self, block: Block) -> AILBlockSplit | None:
        for _, split_blocks in self.original_split_blocks.items():
            for split_block in split_blocks:
                if split_block.original == block:
                    return split_block

        _l.warning("Error in finding split block by original on %s", block)
        return None

    def _find_og_start_by_merge_end(self, merge_end: Block) -> Block | None:
        original = next(iter(self.merge_blocks_to_originals[merge_end]))
        if isinstance(original, AILBlockSplit):
            original = original.original

        for og_start, og_blocks in self.original_blocks.items():
            if original in og_blocks:
                return og_start

        return None

    @staticmethod
    def clone_graph_replace_splits(graph: nx.DiGraph, split_map: dict[Block, Block]):
        # do a deepcopy, so we don't edit the original graph
        graph = copy_graph_and_nodes(graph)

        # replace every block that has been split
        updated_blocks = {}
        for original_node, new_node in split_map.items():
            graph.add_node(new_node)
            # correct every in_edge to this node, with new targets for jumps
            for pred in list(graph.predecessors(original_node)):
                new_pred = pred.copy()
                new_pred.statements[-1] = correct_jump_targets(
                    new_pred.statements[-1], {original_node.addr: new_node.addr}, new_stmt=True
                )
                updated_blocks[pred] = new_pred
                replace_node_in_graph(graph, pred, new_pred)
                graph.add_edge(new_pred, new_node)

            # re-add every out_edge
            blk = updated_blocks.get(original_node, original_node)
            for succ in graph.successors(blk):
                graph.add_edge(new_node, succ)

            # finally, kill the original
            if original_node in graph:
                graph.remove_node(original_node)

        return graph, updated_blocks



from __future__ import annotations
from .base import SimSootValue
from angr.engines.soot.field_dispatcher import resolve_field


class SimSootValue_StaticFieldRef(SimSootValue):
    __slots__ = ["class_name", "field_name", "id", "type"]

    def __init__(self, class_name, field_name, type_):
        self.id = f"{class_name}.{field_name}"
        self.class_name = class_name
        self.field_name = field_name
        self.type = type_

    def __repr__(self):
        return self.id

    @classmethod
    def from_field_id(cls, field_id):
        return cls(field_id.class_name, field_id.name, field_id.type)

    @classmethod
    def from_sootvalue(cls, soot_value, state):
        field_name, field_class_name = soot_value.field
        field_type = soot_value.type
        # return field reference
        return cls.get_ref(state, field_class_name, field_name, field_type)

    @classmethod
    def get_ref(cls, state, field_class_name, field_name, field_type):
        """
        Resolve the field within the given state.
        """
        # resolve field
        field_class = state.javavm_classloader.get_class(field_class_name)
        field_id = resolve_field(state, field_class, field_name, field_type)
        # return field ref
        return cls.from_field_id(field_id)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations
from angr.ailment.expression import BinaryOp, Const, Load

from angr.utils.loader import is_pc, is_in_readonly_section, is_in_readonly_segment
from .base import PeepholeOptimizationExprBase


class SimplifyPcRelativeLoads(PeepholeOptimizationExprBase):
    """
    Simplifying pc-relative loads.
    """

    __slots__ = ()

    NAME = "Simplify PC-relative loads"
    expr_classes = (BinaryOp,)

    def optimize(self, expr: BinaryOp, **kwargs):
        # Load(addr) + pc ==> Const()
        if expr.op == "Add" and len(expr.operands) == 2 and isinstance(expr.operands[0], Load):
            op0, op1 = expr.operands

            assert self.project is not None
            if "ins_addr" not in expr.tags:
                return expr
            assert expr.tags["ins_addr"] is not None

            # check if op1 is PC
            if (
                isinstance(op1, Const)
                and op1.is_int
                and "ins_addr" in expr.tags
                and is_pc(self.project, expr.tags["ins_addr"], op1.value)  # type: ignore
                and isinstance(op0.addr, Const)
                and op0.addr.is_int
            ):
                # check if op0.addr points to a read-only section
                addr = op0.addr.value
                assert isinstance(addr, int)
                if is_in_readonly_section(self.project, addr) or is_in_readonly_segment(self.project, addr):
                    # found it!
                    # do the load first
                    try:
                        offset = self.project.loader.memory.unpack_word(addr, size=self.project.arch.bytes)
                    except KeyError:
                        return expr
                    value = offset + op1.value
                    return Const(None, None, value, self.project.arch.bits, **expr.tags)

        return expr



from __future__ import annotations

from angr.ailment.statement import DirtyStatement, Statement, Call
from angr.ailment.expression import Const, DirtyExpression, Expression
from .rewriter_base import DirtyRewriterBase


class AMD64DirtyRewriter(DirtyRewriterBase):
    """
    Rewrites AMD64 DirtyStatement and DirtyExpression.
    """

    __slots__ = ()

    def _rewrite_stmt(self, dirty: DirtyStatement) -> Statement | None:
        # TODO: Rewrite more dirty statements
        return self._rewrite_expr_to_call(dirty.dirty)

    def _rewrite_expr(self, dirty: DirtyExpression) -> Expression | None:
        return self._rewrite_expr_to_call(dirty)

    def _rewrite_expr_to_call(self, dirty: DirtyExpression) -> Call | None:
        match dirty.callee:
            case "amd64g_dirtyhelper_IN":
                if len(dirty.operands) != 2:
                    return None
                portno, size = dirty.operands
                if not isinstance(size, Const):
                    return None
                bits = size.value_int * self.arch.byte_width
                return Call(
                    idx=dirty.idx,
                    target=f"__in{self._inout_intrinsic_suffix(bits)}",
                    calling_convention=None,
                    prototype=None,
                    args=(portno,),
                    ret_expr=None,
                    bits=dirty.bits,
                    **dirty.tags,
                )
            case "amd64g_dirtyhelper_OUT":
                if len(dirty.operands) != 3:
                    return None
                portno, data, size = dirty.operands
                if not isinstance(size, Const):
                    return None
                bits = size.value_int * self.arch.byte_width
                return Call(
                    dirty.idx,
                    target=f"__out{self._inout_intrinsic_suffix(bits)}",
                    calling_convention=None,
                    prototype=None,
                    args=(portno, data),
                    ret_expr=None,
                    bits=None,
                    **dirty.tags,
                )
        return None

    #
    # in, out
    #

    @staticmethod
    def _inout_intrinsic_suffix(bits: int) -> str:
        match bits:
            case 8:
                return "byte"
            case 16:
                return "word"
            case 32:
                return "dword"
            case _:
                return f"_{bits}"



from __future__ import annotations
import logging

import claripy

import angr

l = logging.getLogger(name=__name__)


class lseek(angr.SimProcedure):
    def run(self, fd, seek, whence):  # pylint:disable=arguments-differ,unused-argument
        if self.state.solver.symbolic(whence):
            err = "Symbolic whence is not supported in lseek syscall."
            l.error(err)
            raise angr.errors.SimPosixError(err)

        whence = self.state.solver.eval(whence)
        if whence == 0:
            whence_str = "start"
        elif whence == 1:
            whence_str = "current"
        elif whence == 2:
            whence_str = "end"
        else:
            return -1

        # let's see what happens...
        # if self.state.solver.symbolic(seek):
        #    err = "Symbolic seek is not supported in lseek syscall."
        #    l.error(err)
        #    raise angr.errors.SimPosixError(err)

        # seek = self.state.solver.eval(seek)

        simfd = self.state.posix.get_fd(fd)
        if simfd is None:
            return -1
        success = simfd.seek(seek, whence_str)
        if self.state.solver.is_false(success):
            return -1
        return claripy.If(success, simfd.tell(), -1)
